///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: AR-DS-StreamInfo.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which encapsulates all of the details
//   of an audio stream that is being played by the DirectSound based
//   audio renderer.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "AR-DS-StreamInfo.h"
#include "dx-errors.h"
#include "trace.h"


// Define this to have a copy of all data sent to the sound card
// also written to disk.
//#define ALSO_WRITE_TO_DISK


// ===============================================================================

//
// StreamInfo::StreamInfo
//
// Constructor, set default values.
//

StreamInfo::StreamInfo(LPDIRECTSOUND pDS, AudioSourceProxy* pSrc,
					   DWORD startTime)
{
	m_pDS = pDS;
	m_pDSBuf = NULL;
	m_startTime = startTime;
	m_pSrc = pSrc;
	m_nextWrite = 0;
	m_playing = FALSE;

	// This signifies that the audio format is not known
	m_wf.nChannels = 0;

	m_rawOutput = NULL;
#ifdef ALSO_WRITE_TO_DISK
	m_rawOutput = fopen("samplesPlayed.raw","wb");
	assert(m_rawOutput != NULL);
#endif
} /* StreamInfo::StreamInfo */


// ===============================================================================

//
// StreamInfo::~StreamInfo
//
// Destructor, shut down the stream.
//

StreamInfo::~StreamInfo()
{
	if (m_pDSBuf != NULL)
	{
		m_pDSBuf->Stop();
		m_pDSBuf->Release();
	}

	if (m_rawOutput != NULL)
		fclose(m_rawOutput);
} /* StreamInfo::~StreamInfo */


// ===============================================================================


//
// StreamInfo::Create
//
// Create a streaming buffer.  Return DS_OK if created OK.
//

int StreamInfo::Create(DWORD nSamplesPerSec, DWORD nBitsPerSample,
					   WORD nNumChannels)
{
	// Specify the wave format
	memset(&m_wf, 0, sizeof(WAVEFORMATEX));
	m_wf.wFormatTag      = WAVE_FORMAT_PCM; 
	m_wf.nChannels       = nNumChannels;
	m_wf.nSamplesPerSec  = nSamplesPerSec;
	m_wf.wBitsPerSample  = nBitsPerSample;
	m_wf.nBlockAlign     = m_wf.nChannels * m_wf.wBitsPerSample / 8;
	m_wf.nAvgBytesPerSec = nSamplesPerSec * m_wf.nBlockAlign; 
	m_wf.cbSize          = 0;

	// Set up DSBUFFERDESC structure.
	memset(&m_bufDesc, 0, sizeof(DSBUFFERDESC));
	m_bufDesc.dwSize = sizeof(DSBUFFERDESC);
	m_bufDesc.dwFlags = DSBCAPS_CTRLDEFAULT; 
	m_bufDesc.dwBufferBytes = m_wf.nAvgBytesPerSec * BUFFER_LENGTH / 1000;
	m_bufDesc.lpwfxFormat = &m_wf;

	// Create the DirectSound buffer
	HRESULT res = m_pDS->CreateSoundBuffer(&m_bufDesc, &m_pDSBuf, NULL);

	return res;
} /* StreamInfo::Create */


// ===============================================================================

//
// StreamInfo::GetMaxWriteSize
//
// Calculate the maximum size of sound buffer data which can be written
// to the given stream.  i.e. how much free space there is in buffer.
//

DWORD StreamInfo::GetMaxWriteSize()
{
    DWORD dwWriteCursor, dwPlayCursor, dwMaxSize;

    // Get current play position
    if (m_pDSBuf->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor) == DS_OK)
    {
		// Handle the special case at the start, when the buffer is
		// completely empty, and the stream is not playing
        if ((m_nextWrite == dwPlayCursor) && (!IsPlaying()))
			dwMaxSize = m_bufDesc.dwBufferBytes;

		else if (m_nextWrite <= dwPlayCursor)
		{
			// Our write position trails play cursor
			dwMaxSize = dwPlayCursor - m_nextWrite;
		}

		else // (m_nextWrite > dwPlayCursor)
		{
			// Play cursor has wrapped
			dwMaxSize = m_bufDesc.dwBufferBytes - m_nextWrite + dwPlayCursor;
		}

		TRACE_PRINT3("nextWrite = %d, playCursor = %d, maxSize = %d\n",
					 m_nextWrite, dwPlayCursor, dwMaxSize);
    }
    else
    {
        // GetCurrentPosition call failed
        assert(0);
        dwMaxSize = 0;
    }

    return (dwMaxSize);
} /* StreamInfo::GetMaxWriteSize */


// ===============================================================================

//
// StreamInfo::WriteData
//
// Write the given data to the stream buffer.
//

void StreamInfo::WriteData(LPBYTE pData, DWORD nNumBytes)
{
    HRESULT hr;
    LPBYTE lpBuf1 = NULL;
    LPBYTE lpBuf2 = NULL;
    DWORD dwSize1 = 0;
    DWORD dwSize2 = 0;

#ifdef ALSO_WRITE_TO_DISK
	fwrite(pData, nNumBytes, 1, m_rawOutput);
#endif

    // Lock the sound buffer
    hr = m_pDSBuf->Lock(m_nextWrite, nNumBytes, &lpBuf1, &dwSize1,
						&lpBuf2, &dwSize2, 0);
    assert(hr == DS_OK);

    // Write data to sound buffer. Because the sound buffer is
	// circular, we may have to do two write operations if the locked
	// portion of buffer wraps around to start of buffer.
    assert(lpBuf1 != NULL);
	memcpy(lpBuf1, pData, dwSize1);

	// Second write required?
    if (lpBuf2 != NULL)
    {
		pData += dwSize1;
		memcpy(lpBuf2, pData, dwSize2);
    }

    // Update the buffer offset and unlock sound buffer
    m_nextWrite = (m_nextWrite + nNumBytes) % m_bufDesc.dwBufferBytes;
    m_pDSBuf->Unlock(lpBuf1, dwSize1, lpBuf2, dwSize2);
} /* StreamInfo::WriteData */


// ===============================================================================

//
// StreamInfo::WriteSilence
//
// Write the given number of bytes worth of silence into the buffer.
//

void StreamInfo::WriteSilence(DWORD nNumBytes)
{
    HRESULT hr;
    LPBYTE lpBuf1 = NULL;
    LPBYTE lpBuf2 = NULL;
    DWORD dwSize1 = 0;
    DWORD dwSize2 = 0;

#ifdef ALSO_WRITE_TO_DISK
	for (DWORD i = 0; i < nNumBytes; i++)
		fputc(0, m_rawOutput);
#endif

    // Lock the sound buffer
    hr = m_pDSBuf->Lock(m_nextWrite, nNumBytes, &lpBuf1, &dwSize1,
						&lpBuf2, &dwSize2, 0);
    assert(hr == DS_OK);

    // Write data to sound buffer. Because the sound buffer is
	// circular, we may have to do two write operations if the locked
	// portion of buffer wraps around to start of buffer.
    assert(lpBuf1 != NULL);
	if (m_wf.wBitsPerSample == 8)
		memset(lpBuf1, 0, dwSize1);
	else
	{
		assert((dwSize1 % sizeof(signed short)) == 0);
		signed short* data = (signed short*) lpBuf1;
		for (unsigned int i = 0; i < dwSize1 / sizeof(signed short); i++)
			data[i] = 0;
	}

	// Second write required?
    if (lpBuf2 != NULL)
	{
		if (m_wf.wBitsPerSample == 8)
			memset(lpBuf2, 0, dwSize2);
		else
		{
			assert((dwSize2 % sizeof(signed short)) == 0);
			signed short* data = (signed short*) lpBuf2;
			for (unsigned int i = 0; i < dwSize2 / sizeof(signed short); i++)
				data[i] = 0;
		}
	}

    // Update the buffer offset and unlock sound buffer
    m_nextWrite = (m_nextWrite + nNumBytes) % m_bufDesc.dwBufferBytes;
    m_pDSBuf->Unlock(lpBuf1, dwSize1, lpBuf2, dwSize2);
} /* StreamInfo::WriteSilence */


// ===============================================================================

//
// StreamInfo::Play
//
// Start the stream playing.
//

void StreamInfo::Play()
{
	HRESULT hr = m_pDSBuf->Play(0, 0, DSBPLAY_LOOPING);
	assert(hr == DS_OK);
	if (hr == DS_OK)
		m_playing = TRUE;
} /* StreamInfo::Play */


// ===============================================================================
