///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  AudioSourceProxy.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the AudioSource node.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "AudioSourceProxy.h"
#include "Clock.h"
#include "trace.h"


//FILE* outfile;


// =====================================================================

//
// PresenterCSELT::InitAudioSource
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitAudioSource(AudioSource* pObj)
{
	pObj->SetPresenterData(new AudioSourceProxy(pObj));
}


// ======================================================================

//
// AudioSourceProxy::AudioSourceProxy
//
// Class constructor.
//

AudioSourceProxy::AudioSourceProxy(MediaObject* pMO) : Traversable(pMO)
{
	m_registeredWithRenderer = FALSE;
	m_skippedHeader = FALSE;
	m_partialStoreAllocatedSize = 0;
	m_partialStore = NULL;
	m_partialLen = 0;
	m_needsReleasing = TRUE;
	m_isFinished = FALSE;
	m_firstTimeAtStartTime = TRUE;
	m_firstTimeAtStopTime  = TRUE;
	m_timesStarted = 0;
} /* AudioSourceProxy::AudioSourceProxy */


// ======================================================================

//
// AudioSourceProxy::~AudioSourceProxy
//
// Class destructor.
//

AudioSourceProxy::~AudioSourceProxy()
{
	//fclose(outfile);
	if (m_registeredWithRenderer)
		GetAudioRenderer()->RemoveSourceNode(this);

	if (m_partialStore)
		delete m_partialStore;
} /* AudioSourceProxy::~AudioSourceProxy */


// =====================================================================

//
// AudioSourceProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void AudioSourceProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "AudioSource {\n");
	const char* indentStr = IndentPrefix(depth + 1);
	fprintf(outfile, "%sobjectDescriptorID %d\n", indentStr,
			(int)GetAudioSource()->GetObjectDescriptorID());
	fprintf(outfile, "%spitch %.3f\n", indentStr,
			(float)GetAudioSource()->pitch);
	fprintf(outfile, "%sstartTime %.2lf\n", indentStr,
			(double)(Time)GetAudioSource()->startTime);
	fprintf(outfile, "%sstopTime %.2lf\n", indentStr,
			(double)(Time)GetAudioSource()->stopTime);
	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* AudioSourceProxy::Print */


// ======================================================================

//
// AudioSourceProxy::PreRender
//
// The first time that this is called, the output of the decoder
// will be read to determine the format of the audio, and then
// register this node with the audio renderer.  This method also
// supports the startTime field.
//

void AudioSourceProxy::PreRender(Effects const& /*effects*/)
{
	// This gets the time since the start of the scene
	// (convert the time to seconds, as used by BIFS)
	double currentTime = (double)GetPresenter()->GetTime() / 1000;

	// Get the current values of the start and stop time from the
	// AudioSource node
	double startTime = (double)GetAudioSource()->startTime;
	double stopTime  = (double)GetAudioSource()->stopTime;

	// Only proceed once a stream has been attached
//	if (GetStream() == NULL)
//		return;

	// Determine if the stop time has been reached
	if ((currentTime >= stopTime) && (stopTime >= startTime))
	{
		// If this is the first time, and the stream is currently
		// not active, then close the stream and remove this node
		// from the audio renderer.
		if ((m_firstTimeAtStopTime) && (m_registeredWithRenderer))
		{
			TRACE_PRINT1("AudioSourceProxy: Stopping at %lf\n", currentTime);
			m_firstTimeAtStopTime = FALSE;
			m_firstTimeAtStartTime = TRUE;

			if (GetAudioSource()->GetStream() != NULL)
			{
				GetAudioSource()->GetStream()->Terminate();
				GetAudioSource()->CloseStream();
			}
			m_isFinished = TRUE;

			if (m_registeredWithRenderer)
			{
				TRACE_PRINT("AudioSourceProxy: removing from audio renderer\n");
				GetAudioRenderer()->RemoveSourceNode(this);
				m_registeredWithRenderer = FALSE;
			}
		}
	}

	// Only proceed if the start time has been reached
	if ((currentTime < startTime) || (stopTime > startTime))
		return;

	// If this is the first time that the start time has been reached
	if (m_firstTimeAtStartTime)
	{
		m_firstTimeAtStartTime = FALSE;
		m_firstTimeAtStopTime = TRUE;
		m_timesStarted++;

		// If this is not the first time that this has been
		// started, we need to reconnect this stream to the demux
		AudioSource* pAS = GetAudioSource();
		if (m_timesStarted > 1)
			pAS->SetStream(pAS->m_copyOfESdesc, pAS->m_copyOfFileName);

		// Start the stream flowing.
		pAS->OpenStream();

		// Read the output of the decoder to get the audio format
		int dataSize;
		DWORD rdwTime;
	
		LPVOID dataPtr = GetStream()->FetchNext(dataSize, rdwTime, TRUE);

		// If the data couldn't be obtained, then don't proceed
		if ((dataPtr == NULL) || (dataPtr == END_OF_STREAM) || (dataSize == 0))
			return;

		// The audio format is passed by the following structure
		DecoderAudioFormat* pFmt = (DecoderAudioFormat*)dataPtr;

		// Record when we started playing (in msec)
		m_startUpDelay = 1000 * startTime;
		TRACE_PRINT1("AudioSourceProxy: Starting at %d\n", m_startUpDelay);

		// Tell the audio renderer about this node
		m_registeredWithRenderer = TRUE;
		GetAudioRenderer()->AddSourceNode(this, startTime,
										  pFmt->samplesPerSecond,
										  pFmt->bitsPerSample,
										  pFmt->numChannels);

		// Release the media stream buffer (but don't discard)
		GetStream()->Release(FALSE);
	}

} /* AudioSourceProxy::PreRender */


// ============================================================================

//
// AudioSourceProxy::GetData
//
// This function is used by the audio renderer to get more data
// to play.  The audio render will ask for a given number of bytes
// (which will be a multiple of the number of bytes required for a
// single sample), and the audio source node will return a pointer
// to some data, and the number of bytes of data pointed to.
//
// The number of bytes returned can be less than that asked for.
// If fewer bytes were returned to the audio renderer than expected,
// then the audio renderer will ask for more.  If zero bytes are
// returned, then the audio renderer will assume that no data can be
// given, and will insert silence.
//

void AudioSourceProxy::GetData(DWORD nBytesRequired, LPBYTE& rpData,
							   DWORD& rnBytesGiven)
{
	// If the stream has already been closed down, then return zero bytes.
	if (GetStream() == NULL)
	{
		rpData = NULL;
		rnBytesGiven = 0;
		return;
	}

	int dataSize;
	DWORD rdwTime;
	BOOL bRecurrent;
	BOOL bEndOfStream;
	LPVOID dataPtr;

	// If there is partial data left over from a previous call, use it
	// instead of reading from the stream
	if (m_partialLen > 0)
	{
		dataSize		 = m_partialLen;
		m_partialLen	 = 0;
		rdwTime			 = m_partialTime;
		bRecurrent		 = FALSE;
		bEndOfStream	 = FALSE;
		dataPtr			 = m_partialStore;
		m_needsReleasing = FALSE;
	}
	else
	{
		// Read the output of the decoder to get the audio format
		dataPtr = GetStream()->Fetch(dataSize, rdwTime, bRecurrent,
									 bEndOfStream, FALSE, FALSE, 300);
		m_needsReleasing = TRUE;
	}

	// If end of stream has been reached, close the stream and return
	// zero bytes
	if (bEndOfStream)
	{
		TRACE_PRINT("AudioSourceProxy::GetData(): End of stream detected.\n");

		GetAudioSource()->GetStream()->Terminate();
		GetAudioSource()->CloseStream();
		rpData = NULL;
		rnBytesGiven = 0;
		m_isFinished = TRUE;
	}

	// If there is no data, or the data returned has already been played,
	// return zero bytes
	else if ((dataPtr == NULL) || (bRecurrent))
	{
		rpData = NULL;
		rnBytesGiven = 0;
	}

	// If data block returned has 0 size, return zero bytes
	else if (dataSize == 0)
	{
		GetStream()->Release();
		rpData = NULL;
		rnBytesGiven = 0;
	}

	else
	{
		// Skip the header data if this is the first time data has been
		// read from this buffer by this function
		if (!m_skippedHeader)
		{
			m_skippedHeader = TRUE;
			rpData = (LPBYTE)dataPtr + sizeof(DecoderAudioFormat);
			rnBytesGiven = dataSize - sizeof(DecoderAudioFormat);
		}

		else
		{
			rpData = (LPBYTE)dataPtr;
			rnBytesGiven = dataSize;
		}

		TRACE_PRINT2("AudioSourceProxy::GetData(): Playing time %ld at %ld\n", rdwTime,
					 GetAudioSource()->GetExecutive()->GetClock()->GetTime());
	}

	// If the number of bytes requested is less than the number read
	// from the media stream, only return some, and then store the
	// rest to use on the next call.
	if (rnBytesGiven > nBytesRequired)
	{
		m_partialLen = rnBytesGiven - nBytesRequired;
		m_partialTime = rdwTime;

		// Make the partial store large enough
		if (m_partialStoreAllocatedSize < m_partialLen)
		{
			if (m_partialStore)
				delete m_partialStore;
			m_partialStore = new unsigned char [m_partialLen];
			m_partialStoreAllocatedSize = m_partialLen;
		}

		// Copy over the remaining samples
		memcpy(m_partialStore, rpData + nBytesRequired, m_partialLen);
		rnBytesGiven = nBytesRequired;
	}
	else
		m_partialLen = 0;

	//if (rpData != NULL)
	//	fwrite(rpData,rnBytesGiven,1,outfile);
} /* AudioSourceProxy::GetData */


// ============================================================================

//
// AudioSourceProxy::DataUsed
//
// This function is used by the audio renderer to indicate that it
// has finished with some data given to it by GetData().  The
// parameters given to this function will be the same as those
// returned by GetData().
//

void AudioSourceProxy::DataUsed(LPBYTE /*pData*/, DWORD /*nBytesGiven*/)
{
	if (m_needsReleasing)
		GetStream()->Release();
} /* AudioSourceProxy::DataUsed */


// ============================================================================
