///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  BRect.h
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The definition of a class which stores bounding rectangle
//   information, and provides methods for operating on bounding
//   rectangles.
//
//   Note that this rectangle uses BIFS coordinates, with the origin
//   in the middle of the screen, increasing X to the right, and
//   increasing Y upwards.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#ifndef BRECT_H
#define BRECT_H


#include <ostream.h>
#include "types-CSELT.h"


class BRect
{
public:
	int left, top, width, height;

	// The default constructor creates an empty bounding box
	BRect() { MakeEmpty(); }

	// Set the dimensions of the bounding rectangle, given the 4 separate
	// values.
	BRect(int l, int t, int w, int h)	{ SetDimensions(l, t, w, h); }
	void SetDimensions(int l, int t, int w, int h);

	// Set the dimensions of the bounding rectangle, given another
	// bounding rectangle.
	BRect(BRect const& other)			{ SetDimensions(other); }
	void SetDimensions(BRect const& other);

	// Set the given bounding rectangle to be the smallest rectangle
	// which completely enclosed the given list of points.
	void SetDimensions(PointF const* lpPoints, int numPoints);

	// Set this bounding rectangle to the dimensions of the given rect.
	void operator= (const BRect& rect) { SetDimensions(rect); }

	// Make this bounding box empty
	void MakeEmpty();

	// Return TRUE if this bounding box is empty, FALSE otherwise
	BOOL IsEmpty() const
	{ return ((width == 0) || (height == 0)); }

	// The current bounding box becomes the union of itself and
	// the given bounding box.
	//  - if other box is empty, nothing happens.
	//  - if this box is empty, then it is set to the other box.
	void Union(BRect const& other);
	void Union(BRect const* pOther) { Union(*pOther); }

	// The current bounding box becomes the intersection of itself and
	// the given bounding box.
	void Intersection(BRect const& other);
	void Intersection(BRect const* pOther) { Intersection(*pOther); }

	// Return TRUE if the current bounding box and the given bounding
	// box overlap, otherwise return FALSE.  If either box is empty,
	// return FALSE.
	BOOL Overlaps(BRect const& other) const;
	BOOL Overlaps(BRect const* pOther) const { return Overlaps(*pOther); }

	// Print out the bounding rectangle
	friend ostream& operator<<(ostream& out, BRect const& bRect);
	void Print(FILE* outfile) const;

	// Return a pointer to a string (in static memory) which describes
	// the dimensions of the bounding rectangle.
	operator char*() const;

	// Return TRUE if the given point is contained in the bounding
	// rectangle, otherwise return FALSE.
	BOOL IsPointIn(int x, int y) const;

	// Return TRUE if the two given bounding rectangles are the same.
	friend int operator== (const BRect& rect1,	const BRect& rect2);

	// Return TRUE if the two given bounding rectangles are different.
	friend int operator!= (const BRect& rect1,	const BRect& rect2)
	{ return !operator==(rect1, rect2); }
};


#endif // BRECT_H
