///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  CircleProxy.cpp
//
// Author: Antonio Arena, CSELT, arena@kim.cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Circle node.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "CircleProxy.h"
#include "trace.h"
#include <assert.h>

// =====================================================================

//
// PresenterCSELT::InitCircle
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitCircle(Circle* pObj)
{
	pObj->SetPresenterData(new CircleProxy(pObj));
}


// ======================================================================

//
// CircleProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void CircleProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "Circle {\n");
	fprintf(outfile, "%sradius %.3f\n", IndentPrefix(depth + 1),
			(float)GetCircle()->radius);
	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* CircleProxy::Print */

// ======================================================================

//
// CircleProxy::PreRender
//
//

void CircleProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, or it is not filled and
	// the line width is 0, then it doesn't need to be drawn, and its
	// bounding rectangle is empty.
	if ((GetTransparency() > 1 - ALMOST_ZERO) ||
		((GetFilled() == FALSE) && (GetWidth() == 0)))
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	assert(!m_transform.ContainsRotation());

	// Put the center always in the origin
	// and then apply the parent's transformations
	// which also determine the circle's bounding box.
	int radius = (float)GetCircle()->radius;
	m_lpPoints[0].x = m_lpPoints[1].y = -radius;
	m_lpPoints[0].y = m_lpPoints[1].x = radius;

	// Find out where the shadow should be drawn
	int i;
	for (i = 0; i < 2; i++)
	{
		m_shadows[i].x = m_lpPoints[i].x + GetShadPosX();
		m_shadows[i].y = m_lpPoints[i].y + GetShadPosY();
	}

	// Find out where the circle should be drawn
	for (i = 0; i< 2; i++)
	{
		m_transform.Apply(m_lpPoints[i].x, m_lpPoints[i].y);
		m_transform.Apply(m_shadows[i].x, m_shadows[i].y);
	}

	m_boundingRect.SetDimensions(m_lpPoints, 2);

	if (GetShadow())
	{
		BRect shadowRect;
		shadowRect.SetDimensions(m_shadows, 2);
		m_boundingRect.Union(shadowRect);
	}

	UpdateBoundingRect();
} /* CircleProxy::PreRender */


// ============================================================================

//
// CircleProxy::Draw
//
//

void CircleProxy::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);

	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(GetBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	// If there's the shadow, draw it now
	if (GetShadow())
	{
		// Default shadow color
		Color3f shadowColor(0, 0.5, 0);

		GetVisualRenderer()->DisplayCircle(visibleRect, m_shadows, shadowColor, 
											GetWidth(), GetFilled());

	}

	// Now draw the circle
	GetVisualRenderer()->DisplayCircle(visibleRect, m_lpPoints, GetDiffuseColor(), 
		GetWidth(), GetFilled());
} /* CircleProxy::Draw */


// ===========================================================================
