///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  Curve2DProxy.cpp
//
// Author: Antonio Arena, CSELT, arena@kim.cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Curve node.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "Curve2DProxy.h"
#include "Coordinate2DProxy.h"
#include "trace.h"
#include <assert.h>

// =====================================================================

//
// PresenterCSELT::InitCurve2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitCurve2D(Curve2D* pObj)
{
	pObj->SetPresenterData(new Curve2DProxy(pObj));
}


// ======================================================================

//
// Curve2DProxy::Curve2DProxy
//
// Class constructor.
//

Curve2DProxy::Curve2DProxy(MediaObject* pMO) : GeometryNode(pMO)
{
	m_lpPoints = NULL;
	m_shadows  = NULL;
	m_nAllocatedLength = 0;
} /* Curve2DProxy::Curve2DProxy */


// ======================================================================

//
// Curve2DProxy::~Curve2DProxy
//
// Class destructor .
//

Curve2DProxy::~Curve2DProxy()
{
	if (m_lpPoints != NULL)	delete [] m_lpPoints;
	if (m_shadows != NULL)	delete [] m_shadows;
} /* Curve2DProxy::~Curve2DProxy */


// ======================================================================

//
// Curve2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void Curve2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "Curve2D {\n");

	fprintf(outfile, "%spoints ", IndentPrefix(depth + 1));
	if (GetCoordinate2D() == NULL)
		fprintf(outfile, "[]\n");
	else
		((Coordinate2DProxy*)(GetCoordinate2D()->GetPresenterData()))->Print(outfile, depth + 1);

	fprintf(outfile, "%sfineness %d\n", IndentPrefix(depth + 1),
			(int)GetCurve()->fineness);

	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* Curve2DProxy::Print */


// ======================================================================

//
// Curve2DProxy::PreRender
//
//

void Curve2DProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, or it is not filled and
	// the line width is 0, then it doesn't need to be drawn, and its
	// bounding rectangle is empty.
	if ((GetTransparency() > 1 - ALMOST_ZERO) ||
		((GetFilled() == FALSE) && (GetWidth() == 0)))
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Allocate memory for the local copy of the points (only if the
	// current array is not large enough).
	m_nLength = GetCoordinate2D()->point.GetLength();
	if (m_nLength > m_nAllocatedLength)
	{
		// Delete the previous arrays - if required
		if (m_lpPoints != NULL)	delete [] m_lpPoints;
		if (m_shadows != NULL)	delete [] m_shadows;

		// Create the new arrays
		m_lpPoints = new PointF[m_nLength];
		m_shadows = new PointF[m_nLength];
		m_nAllocatedLength = m_nLength;
	}

	// Find out where the points and the shadow lie
	for (int i =0; i < m_nLength; i++)
	{
		m_lpPoints[i].x = GetCoordinate2D()->point[i].x; 
		m_lpPoints[i].y = GetCoordinate2D()->point[i].y;
		m_shadows[i].x = m_lpPoints[i].x + GetShadPosX(); 
		m_shadows[i].y = m_lpPoints[i].y + GetShadPosY();
	}

	// Apply the parent's transformations
	// to both the points and the shadows
	for (i =0; i < m_nLength; i++)
	{
		m_transform.Apply(m_lpPoints[i].x, m_lpPoints[i].y);
		m_transform.Apply(m_shadows[i].x, m_shadows[i].y);
	}

	m_boundingRect.SetDimensions(m_lpPoints, m_nLength);

	// If there's the shadow,
	// enlarge the bounding box
	if (GetShadow())
	{
		BRect shadowRect;
		shadowRect.SetDimensions(m_shadows, m_nLength);
		m_boundingRect.Union(shadowRect);
	}

	UpdateBoundingRect();
} /* Curve2DProxy::PreRender */


// ============================================================================

//
// Curve2DProxy::Draw
//
//

void Curve2DProxy::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);

	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(GetBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	// If there's a shadow to be drawn
	// Draw it now
	if (GetShadow())
	{
		Color3f shadowColor(0, 0.5, 0);

		GetVisualRenderer()->DisplayCurve(visibleRect, 
										  m_shadows, 
										  shadowColor, 
									      GetWidth(), 
										  m_nLength);

	}

	// Draw the curve
	GetVisualRenderer()->DisplayCurve(visibleRect, 
									  m_lpPoints, 
									  GetDiffuseColor(), 
									  GetWidth(), 
									  m_nLength);
} /* Curve2DProxy::Draw */


// ===========================================================================
