///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  GeometryNode.h
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which contains details specific to
//   geometry nodes.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "GeometryNode.h"
#include "Shape2DProxy.h"


// ========================================================================

//
// GeometryNode::NeedsDrawing
//
// Returns TRUE if this geometry node needs to be drawn (because it
// has changed since it was last drawn).  This will be true if the
// node has been updated (via BIFS or ROUTES), or the appearance node
// associated with this node has changed, or the transformation has
// changed.
//

BOOL GeometryNode::NeedsDrawing() const
{
	DWORD presTime = GetPresenter()->GetTime();

	// Check if the Appearance node has been updated
	Appearance2D* pAppearance = GetAppearance();
	if (pAppearance != NULL)
	{
		if (pAppearance->GetUpdateTime() >= presTime)
			return TRUE;

		// Check if the material node has been updated
		Material2D* pMaterial = pAppearance->material;
		if ((pMaterial != NULL) && (pMaterial->GetUpdateTime() >= presTime))
			return TRUE;

		// Check if the properties node has been updated
		Properties2D* pProperties = pAppearance->properties;
		if ((pProperties != NULL) && (pProperties->GetUpdateTime() >= presTime))
			return TRUE;
	}

	// Check if the geometry node itself has been updated
	if (GetObject()->GetUpdateTime() >= presTime)
		return TRUE;

	// Check if the transformation has changed
	if (m_transform != m_oldTransform)
		return TRUE;
	else
		return FALSE;
} /* GeometryNode::NeedsDrawing */


// ========================================================================

//
// GeometryNode::FindTouchSensor
//
// Call FindTouchSensor() of this node's enclosing grouping node, which
// is the parent of the Shape2D node that this node is the geometry for.
//

void GeometryNode::FindTouchSensor(EventInfo const* pEvent)
{
	// Go through all of the parents
	int whichParent = 0;
	GroupNode* pParent = GetShape2DProxy()->GetObject()->GetParent(whichParent);
	while (pParent != NULL)
	{
		((Traversable*)pParent->GetPresenterData())->FindTouchSensor(pEvent);
		whichParent++;
		pParent = GetShape2DProxy()->GetObject()->GetParent(whichParent);
	}
} /* GeometryNode::FindTouchSensor */


// =====================================================================

//
// GeometryNode::GetDiffuseColor
//
// Return the 'diffuseColor' field from Material2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Material2D node is not present, then
// the default value is used.
//

Color3f GeometryNode::GetDiffuseColor() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->material == NULL))
		return Color3f(0.8, 0.8, 0.8);
	else
		return (Color3f)m_pAppearance->material->diffuseColor;
} /* GeometryNode::GetDiffuseColor */


// =====================================================================

//
// GeometryNode::GetTransparency
//
// Return the 'transparency' field from Material2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Material2D node is not present, then
// the default value is used.
//

float GeometryNode::GetTransparency() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->material == NULL))
		return 0.0;
	else
		return m_pAppearance->material->transparency;
} /* GeometryNode::GetTransparency */


// =====================================================================

//
// GeometryNode::GetFilled
//
// Return the 'filled' field from Properties2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Properties2D node is not present, then
// the default value is used.
//

Bool GeometryNode::GetFilled() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->properties == NULL))
		return FALSE;
	else
		return (Bool)m_pAppearance->properties->filled;
} /* GeometryNode::GetFilled */


// =====================================================================

//
// GeometryNode::GetShadow
//
// Return the 'shadow' field from Properties2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Properties2D node is not present, then
// the default value is used.
//

Bool GeometryNode::GetShadow() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->properties == NULL))
		return FALSE;
	else
		return (Bool)m_pAppearance->properties->shadow;
} /* GeometryNode::GetShadow */


// =====================================================================

//
// GeometryNode::GetShadPosX
//
// Return the 'shadPos.X' field from Properties2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Properties2D node is not present, then
// the default value is used.
//

float GeometryNode::GetShadPosX() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->properties == NULL))
		return 5;
	else
		return ((Vec2f)m_pAppearance->properties->shadPos).x;
} /* GeometryNode::GetShadPosX */


// =====================================================================

//
// GeometryNode::GetShadPosY
//
// Return the 'shadPos.Y' field from Properties2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Properties2D node is not present, then
// the default value is used.
//

float GeometryNode::GetShadPosY() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->properties == NULL))
		return 5;
	else
		return ((Vec2f)m_pAppearance->properties->shadPos).y;
} /* GeometryNode::GetShadPosY */


// =====================================================================

//
// GeometryNode::GetWidth
//
// Return the 'width' field from Properties2D node in the
// Appearance2D node associated with this geometry object.  If either
// the Appearance2D node or the Properties2D node is not present, then
// the default value is used.
//

int GeometryNode::GetWidth() const
{
	if ((m_pAppearance == NULL) || (m_pAppearance->properties == NULL))
		return 1;
	else
		return (int)m_pAppearance->properties->width;
} /* GeometryNode::GetWidth */


// =====================================================================
