///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  PointSet2DProxy.cpp
//
// Author: Antonio Arena, CSELT, arena@kim.cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to handle the PointSet2DProxy node.
//
// Created: September, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "PointSet2DProxy.h"
#include "Coordinate2DProxy.h"
#include "ColorProxy.h"

// =====================================================================

//
// PresenterCSELT::InitPointSet2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void PresenterCSELT::InitPointSet2D(PointSet2D* pObj)
{
	pObj->SetPresenterData(new PointSet2DProxy(pObj));
}


// ======================================================================

//
// PointSet2DProxy::PointSet2DProxy
//
// Class constructor.
//

PointSet2DProxy::PointSet2DProxy(MediaObject* pMO) : GeometryNode(pMO)
{
	m_lpPoints			= NULL;
	m_shadowPoints		= NULL;
	m_nAllocatedLength	= 0;
} /* PointSet2DProxy::PointSet2DProxy */


// ======================================================================

//
// PointSet2DProxy::~PointSet2DProxy
//
// Class destructor .
//

PointSet2DProxy::~PointSet2DProxy()
{
	if (m_lpPoints != NULL)		delete [] m_lpPoints;
	if (m_shadowPoints != NULL)	delete [] m_shadowPoints;
} /* PointSet2DProxy::~PointSet2DProxy */


// =====================================================================

//
// PointSet2DProxy::Print
//
// Print out the fields and corresponding values for the media object
// derived node that this object is a proxy for, indented according to
// the given depth.
//

void PointSet2DProxy::Print(FILE* outfile, int depth) const
{
	fprintf(outfile, "PointSet2D {\n", IndentPrefix(depth));
	const char* indentStr = IndentPrefix(depth + 1);

	// Print out the list of colours for each point
	fprintf(outfile, "%scolor ", indentStr);
	if (GetColorNode() == NULL)
		fprintf(outfile, "[]\n");
	else
		((ColorProxy*)(GetColorNode()->GetPresenterData()))->Print(outfile, depth + 1);

	// Print out the list of points
	fprintf(outfile, "%scoord ", indentStr);
	if (GetCoordinate2D() == NULL)
		fprintf(outfile, "[]\n");
	else
		((Coordinate2DProxy*)(GetCoordinate2D()->GetPresenterData()))->Print(outfile, depth + 1);

	fprintf(outfile, "%s}\n", IndentPrefix(depth));
} /* PointSet2DProxy::Print */


// =====================================================================

//
// PointSet2DProxy::PreRender
//
// Calculate the bounding rectangle for the set of points, which involves
// calculating where each of the points have been transformed to.
//

void PointSet2DProxy::PreRender(Effects const& effects)
{
	if (!InitPreRender(effects))
		return;

	// If the transparency for this node is 1, then it doesn't need
	// to be drawn, and its bounding rectangle is empty
	if (GetTransparency() > 1 - ALMOST_ZERO)
	{
		m_boundingRect.MakeEmpty();
		UpdateBoundingRect();
		return;
	}

	// Allocate memory for the local copy of the points (only if the
	// current array is not large enough).
	m_nLength = GetCoordinate2D()->point.GetLength();
	if (m_nLength > m_nAllocatedLength)
	{
		// Delete the previous arrays - if required
		if (m_lpPoints != NULL)		delete [] m_lpPoints;
		if (m_shadowPoints != NULL)	delete [] m_shadowPoints;

		// Create the new arrays
		m_lpPoints = new PointF[m_nLength];
		m_shadowPoints = new PointF[m_nLength];
		m_nAllocatedLength = m_nLength;
	}

	// Determine the position of each of the points once the transform
	// matrix has been applied.
	for (int i = 0; i < m_nLength; i++)
	{
		m_lpPoints[i].x = GetCoordinate2D()->point[i].x; 
		m_lpPoints[i].y = GetCoordinate2D()->point[i].y;
		m_transform.Apply(m_lpPoints[i].x, m_lpPoints[i].y);
	}

	// Determine the bounding rectangle for this set of points
	m_boundingRect.SetDimensions(m_lpPoints, m_nLength);

	if (GetShadow())
	{
		// Determine where the shadow points are transformed to
		for (int i = 0; i < m_nLength; i++)
		{
			m_shadowPoints[i].x = GetCoordinate2D()->point[i].x
				+ GetShadPosX();
			m_shadowPoints[i].y = GetCoordinate2D()->point[i].y
				+ GetShadPosY();
			m_transform.Apply(m_shadowPoints[i].x, m_shadowPoints[i].y);
		}

		// Extend the bounding rectangle to include the shadow points
		BRect shadowRect;
		shadowRect.SetDimensions(m_shadowPoints, m_nLength);
		m_boundingRect.Union(shadowRect);
	}

	UpdateBoundingRect();
} /* PointSet2DProxy::PreRender */


// ============================================================================

//
// PointSet2DProxy::Draw
//
// Draw the points which lie within the given clipping rectangle.
// Transparency is supported, along with either the same colour for
// all points, or a different colour for each point.
//

void PointSet2DProxy::Draw(BRect const* pClipBRect)
{
	assert(pClipBRect != NULL);
	VisualRenderer* pVR = GetVisualRenderer();

	// The points to draw lie within both the clipping rectangle and the
	// point set's own bounding rectangle.
	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(GetBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	// If a colour for each point is given, check that there are enough
	// colours specified
	if (GetColorNode() != NULL)
		assert(GetColorNode()->color.GetLength() >= m_nLength);

	// Get the transparency to draw the points with
	float transparency = GetTransparency();

	// The diffuse colour is used if the list of colours is not given
	Color3f diffuseColour = GetDiffuseColor();

	// If a shadow is required, draw it now
	if (GetShadow())
	{
		// Determine where to draw the shadow, allowing for the transformation
		int shadowPosX = GetShadPosX();
		int shadowPosY = GetShadPosY();

		// Only draw the shadow if it is not directly underneath
		if ((shadowPosX != 0) || (shadowPosY != 0))
		{
			Color3f shadowColor(0, 0.5, 0);

			// Draw all of the shadow points
			for (int i = 0; i < m_nLength; i++)
			{
				int x = m_shadowPoints[i].x;
				int y = m_shadowPoints[i].y;
				if (visibleRect.IsPointIn(x, y))
					pVR->DisplayPoint(x, y, &shadowColor, transparency);
			}
		}
	}

	// Go through the list of points, drawing those which are in the visible rect
	for (int i = 0; i < m_nLength; i++)
	{	
		int x = m_lpPoints[i].x;
		int y = m_lpPoints[i].y;
		if (visibleRect.IsPointIn(x, y))
		{
			// Determine the colour to draw this point
			Color3f const* pThisColor;
			if (GetColorNode() != NULL)
				pThisColor = &(GetColorNode()->color[i]);
			else
				pThisColor = &diffuseColour;

			pVR->DisplayPoint(x, y, pThisColor, transparency);
		}
	}
} /* PointSet2DProxy::Draw */


// ===========================================================================
