/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor / Presenter
FILE:       StreamCo.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of StreamConsumer
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "StreamCo.h"
#include "Stream.h"
#include "Executiv.h"
#include "AL.h"
#include "Decoder.h"
#include "JPEGDec.h"
#include "H263Dec.h"
#include "G723Dec.h"
#include "PCMDec.h"
#include "RGBDec.h"
#include "AACDecoder.h"
#include "MPEG4Dec.h"
#include "Caps.h"
#include "BasePresenter.h"
#include "Channel.h"
#include "File.h"

/*-----------------------------------------------------------------
StreamConsumer constructor
-----------------------------------------------------------------*/
StreamConsumer :: StreamConsumer ()
{
	m_pStream = NULL;
	m_pDecoder = NULL;
	m_pChannel = NULL;
	m_pAL = NULL;
	m_pService = NULL;
	m_fmtStreamFormat = FORMAT_UNKNOWN;
	objectDescriptorID = -1;

	m_copyOfESdesc = NULL;
	m_copyOfFileName = NULL;

	CUSTOMIZE_STREAMCONSUMER_CONSTRUCTOR
}

/*-----------------------------------------------------------------
StreamConsumer destructor. First flush out everything from the stream
(to make sure no object is waiting on the stream, then delete the
associated decoder and stream.
-----------------------------------------------------------------*/
StreamConsumer :: ~StreamConsumer ()
{
	CUSTOMIZE_STREAMCONSUMER_DESTRUCTOR

	if (m_copyOfESdesc)
		delete m_copyOfESdesc;
	if (m_copyOfFileName)
		free(m_copyOfFileName);

	if (objectDescriptorID >= 0 && GetExecutive () != NULL)
		GetExecutive () -> UnregisterStreamConsumer (this);
	//ML added this line to enable a decoder to shutdown a stream
	//   which hasn't already been done because it wasn't started.
	if (m_pDecoder)
		m_pDecoder->CloseOpenStreams();

	if (m_pStream)
	{	
		m_pStream -> Terminate ();
		CloseStream ();
	}
	if (m_pDecoder)
	{	m_pDecoder -> Close ();
		delete m_pDecoder;
	}
	delete m_pStream;
	if (m_pService)
	{
		m_pService -> Close ();
		delete m_pService;
	}
	delete m_pAL;
}

/*-----------------------------------------------------------------
Overload Parse(). After parsing in the usual way, the ODid field
is checked. If has value, register with Executive.
-----------------------------------------------------------------*/
BOOL StreamConsumer :: Parse (BIFSDecoder *pBifs)
{
	BOOL bResult = MediaObject :: Parse (pBifs);
	if (objectDescriptorID >= 0)
		GetExecutive () -> RegisterStreamConsumer (this);
	return bResult;
}
 
/*-----------------------------------------------------------------
An object descriptor received which connect a stream to this object.
Instantiate the proper decoder and attach it to the object through
streams
-----------------------------------------------------------------*/
void StreamConsumer :: SetStream (
	ES_DESCRIPTOR *pES_Descriptor,
	LPCSTR szFileName
)
{
	extern ESCONFIGPARAMS g_arPredefined [];
	extern int g_nPredefinedLength;

	//ML - Make a copy of the ES_Descriptor and the file name, which
	//     will be used to reopen this stream later on
	if (m_copyOfESdesc == NULL)
	{
		m_copyOfESdesc = new ES_DESCRIPTOR;
		memcpy((char*)m_copyOfESdesc, (char*)pES_Descriptor, sizeof(ES_DESCRIPTOR));
		strcpy(m_copyOfESdesc->m_szDecType, pES_Descriptor->m_szDecType);
		memcpy((char*)&(m_copyOfESdesc->m_ESConfigParams),
			   (char*)&(pES_Descriptor->m_ESConfigParams),
			   sizeof(ESCONFIGPARAMS));

		m_copyOfFileName = strdup(szFileName);
	}

	//ML - Delete the decoder if one already exists
	if (m_pDecoder)
	{	m_pDecoder -> Close ();
		delete m_pDecoder;
	}

	/*-----------------------------------------------------------------
	Parse the name of the decoder and instantiate the proper one
	-----------------------------------------------------------------*/
	if (!strcmp (pES_Descriptor -> m_szDecType, "visual/JPEG"))
		m_pDecoder = new JPEGDecoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "audio/G723"))
		m_pDecoder = new G723Decoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "audio/PCM"))
		m_pDecoder = new PCMDecoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "visual/H263"))
		m_pDecoder = new H263Decoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "visual/RGB"))
		m_pDecoder = new RGBDecoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "audio/AAC"))
		m_pDecoder = new AACDecoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "visual/MPEG4"))
		m_pDecoder = new MPEG4Decoder (GetExecutive ());
	else if (!strcmp (pES_Descriptor -> m_szDecType, "visual/TEXT"))
	{
		/*-----------------------------------------------------------------
		Text is a special case in which there is no decoder. The input
		streams feeds the object directly.
		-----------------------------------------------------------------*/
		m_fmtStreamFormat = FORMAT_TEXT;
	}
	else
	{
		assert(0);	// Shouldn't get here.
	}

	/*-----------------------------------------------------------------
	Verify that there is a matching capability beteen the decoder and
	the presenter, and sets up the decoder format. If no matching
	capability exist, delete the decoder and ignore the flow
	-----------------------------------------------------------------*/
	if (m_pDecoder)
	{
		Capabilities caps;

		GetExecutive () -> GetPresenter () -> GetCapabilities (&caps);
		m_fmtStreamFormat = m_pDecoder -> SetFormat (&caps);
		if (m_fmtStreamFormat == FORMAT_UNKNOWN)
		{	delete m_pDecoder;
			m_pDecoder = NULL;
		}
	}

	m_configParams = pES_Descriptor -> m_ESConfigParams;

	if (m_configParams.m_nPredefined > 0)
	{
		assert (m_configParams.m_nPredefined < g_nPredefinedLength);
		m_configParams = g_arPredefined [m_configParams.m_nPredefined];
	}
	m_nESNumber = pES_Descriptor -> m_nES_number;

	/*-----------------------------------------------------------------
	If file name was not included in the OD, attach the stream to the
	current ALManager, otherwise, create a new FileService
	-----------------------------------------------------------------*/
	if (szFileName [0] == 0)
		AttachStream (GetExecutive () -> GetALManager ());
	else
	{
		//ML - added the extra lines to delete the service and AL manager
		//     if they already exists
		if (m_pService)
		{
			m_pService -> Close ();
			delete m_pService;
		}
		if (m_pAL)
			delete m_pAL;

		m_pAL = new ALManager (this);
		m_pService = new FileService (m_pAL);
		m_pService -> Connect (szFileName);
	}
}

/*-----------------------------------------------------------------
Attach a stream to the ALManager
-----------------------------------------------------------------*/
void StreamConsumer :: AttachStream (ALManager *pAL)
{
	/*-----------------------------------------------------------------
	If the decoder type is not recoginzed, ignore this stream.
	Demux :: AttachStream() is called with NULL MediaStream, which
	will cause the demux to throw away all data of this channel.
	-----------------------------------------------------------------*/
	if (m_fmtStreamFormat == FORMAT_UNKNOWN)
	{	pAL -> AttachStream (
			(objectDescriptorID << 6) | m_nESNumber,
			NULL,
			&m_configParams
		);
		return;
	}
	
	/*-----------------------------------------------------------------
	Create the streams. The buffer size is set to either this provided
	by the object descriptor, or to the one recommended by the decoder.
	In both cases the buffer size is extended by a one unit size, because
	since allocation on the stream is of blocks of continuous memory,
	it might happen that not all buffer size can be used.
	-----------------------------------------------------------------*/
	MediaStream *m_pMuxStream =
		new MediaStream (
			GetExecutive () -> GetClock (),
			m_configParams. m_nBufferSizeEB ?
			m_configParams. m_nBufferSizeEB :
				m_pDecoder ?
					m_pDecoder -> GetRecommendedInputBufferSize () +
					m_pDecoder -> GetOptimalInputSize () : 512
		);

	m_pStream =
		m_pDecoder ?
			new	MediaStream (
				GetExecutive () -> GetClock (),
				m_pDecoder -> GetRecommendedOutputBufferSize () +
				m_pDecoder -> GetOptimalOutputSize ()
			) : m_pMuxStream;

	/*-----------------------------------------------------------------
	Now kick off the decoder
	-----------------------------------------------------------------*/
	if (m_pDecoder)
		m_pDecoder -> Start (m_pMuxStream, m_pStream);

	/*-----------------------------------------------------------------
	Open the mux channel
	-----------------------------------------------------------------*/
	m_pChannel = pAL -> AttachStream (
		(objectDescriptorID << 6) | m_nESNumber,
		m_pMuxStream,
		&m_configParams,
		m_pDecoder ? m_pDecoder -> GetOptimalInputSize () : 50
	);
}

/*-----------------------------------------------------------------
Request transmission of data on the channel
-----------------------------------------------------------------*/
void StreamConsumer :: OpenStream ()
{
	if (m_pChannel)
		m_pChannel -> ChannelReady ();
	GetExecutive () -> StreamOpen ();
}

/*-----------------------------------------------------------------
Close the stream and notify the executive
-----------------------------------------------------------------*/
void StreamConsumer :: CloseStream ()
{
	if (m_pStream)
	{	delete m_pStream;
		m_pStream = NULL;
		GetExecutive () -> StreamClose ();
	}
}

/*-----------------------------------------------------------------
Connection to an auxiliary service (auxiliary file), completed
-----------------------------------------------------------------*/
void StreamConsumer :: OnConnect (CONNECTION_RESULT result)
{
	if (result == CONNECTION_SUCCESS)
		AttachStream (m_pAL);
}
