///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  TransformMatrix2D.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of a class which implements a 2-D transform
//   matrix.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <stdio.h>
#include <math.h>
#include "TransformMatrix2D.h"


// ======================================================================

//
// TransformMatrix2D::SetIdentity
//
// Set the matrix to the identity matrix.
//

void TransformMatrix2D::SetIdentity()
{
	for (int row = 0; row < 3; row++)
		for (int col = 0; col < 3; col++)
			if (row == col)
				m_data[row][col] = (float)1;
			else
				m_data[row][col] = (float)0;
} /* TransformMatrix2D::SetIdentity */


// ======================================================================

//
// TransformMatrix2D::SetTranslation
//
// Set the matrix to represent the given translation.
//

void TransformMatrix2D::SetTranslation(float dx, float dy)
{
	SetIdentity();
	m_data[0][2] = dx;
	m_data[1][2] = dy;
} /* TransformMatrix2D::SetTranslation */


// ======================================================================

//
// TransformMatrix2D::SetScale
//
// Set the matrix to represent the scale (in X and Y).
//

void TransformMatrix2D::SetScale(float sx, float sy)
{
	SetIdentity();
	m_data[0][0] = sx;
	m_data[1][1] = sy;
} /* TransformMatrix2D::SetScale */


// ======================================================================

//
// TransformMatrix2D::SetRotation
//
// Set the matrix to represent the given anticlockwise rotation
// (in radians).
//

void TransformMatrix2D::SetRotation(float angle)
{
	m_data[0][0] = (float)cos((double)angle);
	m_data[0][1] = -(float)sin((double)angle);
	m_data[0][2] = (float)0;
	m_data[1][0] = (float)sin((double)angle);
	m_data[1][1] = (float)cos((double)angle);
	m_data[1][2] = (float)0;
	m_data[2][0] = (float)0;
	m_data[2][1] = (float)0;
	m_data[2][2] = (float)1;
} /* TransformMatrix2D::SetRotation */


// ======================================================================

//
// TransformMatrix2D::IsOnlyTranslation
//
// Return TRUE if the transform matrix specifies only translation.
//

BOOL TransformMatrix2D::IsOnlyTranslation() const
{
	if ((m_data[0][0] == (float)1) && (m_data[0][1] == (float)0) &&
		(m_data[1][0] == (float)0) && (m_data[1][1] == (float)1))
		return TRUE;
	else
		return FALSE;
} /* TransformMatrix2D::IsOnlyTranslation */


// ======================================================================

//
// TransformMatrix2D::ContainsRotation
//
// Return TRUE if the transform matrix contains some rotation.
//

BOOL TransformMatrix2D::ContainsRotation() const
{
	if ((m_data[0][1] == (float)0) && (m_data[1][0] == (float)0))
		return FALSE;
	else
		return TRUE;
} /* TransformMatrix2D::ContainsRotation */


// ======================================================================

//
// TransformMatrix2D::Apply
//
// Apply the transform to the given point (given two floats).  This
// assumes that the given point is placed in a column vector, and then
// the matrix multiplication is performed.
//

void TransformMatrix2D::Apply(float& x, float& y) const
{
	float newX = m_data[0][0] * x + m_data[0][1] * y + m_data[0][2];
	float newY = m_data[1][0] * x + m_data[1][1] * y + m_data[1][2];
	x = newX;
	y = newY;
} /* TransformMatrix2D::Apply */


// ======================================================================

//
// TransformMatrix2D::Apply
//
// Apply the transform to the given point (given two long integers).  This
// assumes that the given point is placed in a column vector, and then
// the matrix multiplication is performed.
//

void TransformMatrix2D::Apply(long& x, long& y) const
{
	float newX = m_data[0][0] * (float)x + m_data[0][1] * (float)y + m_data[0][2];
	float newY = m_data[1][0] * (float)x + m_data[1][1] * (float)y + m_data[1][2];
	x = (long)(newX + 0.5);
	y = (long)(newY + 0.5);
} /* TransformMatrix2D::Apply */


// ======================================================================

//
// TransformMatrix2D::Apply
//
// Apply the transform to the given point (given two long integers).  This
// assumes that the given point is placed in a column vector, and then
// the matrix multiplication is performed.
//

void TransformMatrix2D::Apply(int& x, int& y) const
{
	float newX = m_data[0][0] * (float)x + m_data[0][1] * (float)y + m_data[0][2];
	float newY = m_data[1][0] * (float)x + m_data[1][1] * (float)y + m_data[1][2];
	x = (int)(newX + 0.5);
	y = (int)(newY + 0.5);
} /* TransformMatrix2D::Apply */


// ======================================================================

//
// operator*(TransformMatrix2D, TransformMatrix2D)
//
// Multiple two matricies together, returning a third.
//

TransformMatrix2D operator* (const TransformMatrix2D& mat1,
							 const TransformMatrix2D& mat2)
{
	TransformMatrix2D result;

	for (int r = 0; r < 3; r++)
		for (int c = 0; c < 3; c++)
		{
			result.m_data[r][c] = (float)0;
			for (int i = 0; i < 3; i++)
				result.m_data[r][c] += mat1.m_data[r][i] * mat2.m_data[i][c];
		}

	return result;
} /* operator*(TransformMatrix2D, TransformMatrix2D) */


// ======================================================================

//
// operator==(TransformMatrix2D, TransformMatrix2D)
//
// Test for equality, returning 1 if the two matricies are the same,
// or 0 if they are different.
//

int operator== (const TransformMatrix2D& mat1,
				const TransformMatrix2D& mat2)
{
	for (int r = 0; r < 3; r++)
		for (int c = 0; c < 3; c++)
			if (mat1.m_data[r][c] != mat2.m_data[r][c])
				return 0;

	return 1;
} /* operator==(TransformMatrix2D, TransformMatrix2D) */


// ======================================================================

//
// operator==(TransformMatrix2D, TransformMatrix2D)
//
// Test for inequality, returning 1 if the two matricies are the same,
// or 0 if they are different.
//

int operator!= (const TransformMatrix2D& mat1,
				const TransformMatrix2D& mat2)
{
	return !operator==(mat1, mat2);
} /* operator!=(TransformMatrix2D, TransformMatrix2D) */


// ======================================================================

//
// TransformMatrix2D::Print
//
// Print out the contents of this matrix (only intended for debugging).
//

void TransformMatrix2D::Print(FILE* outfile) const
{
	for (int row = 0; row < 3; row++)
	{
		for (int col = 0; col < 3; col++)
			fprintf(outfile, "%8.4f ", m_data[row][col]);
		fprintf(outfile,"\n");
	}
} /* TransformMatrix2D::Print */


// ======================================================================

#ifdef MAKE_TEST_APP

int main()
{
	TransformMatrix2D m1, m2, m3;

	m1.SetIdentity();
	m1.Print(stdout);
	printf("\n");

	m2.SetTranslation((float)10, (float)20);
	m2.Print(stdout);
	printf("\n");

	m3 = m2 * m1;
	m3.Print(stdout);
	printf("\n");

	return 0;
} /* main */

#endif // MAKE_TEST_APP

// ======================================================================
