///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: VisRend-CSELT.cpp
//
// Authors: Matthew Leditschke <CSELT, Matthew.Leditschke@cselt.it>
//
// Description:
//   The implementation of an abstract base class which performs visual
//   rendering.  Actual implementations of the visual renderer will
//   derive from this class.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend.h"
#include "DrawableNode.h"
#include "BRect.h"
#include "trace.h"

// Include the details of the Windows 95/NT specific audio renderers
#ifdef WIN32
#include "VisRend-CSELT-DD.h"
#endif

// Include the details of the SGI specific audio renderers
#if sgi
#include "VisRend-CSELT-X11.h"
#endif

// Include the details of the sun specific audio renderers
#if sun
#endif


// ======================================================================

//
// VisualRenderer::Create
//
// Return a pointer to a visual renderer.  Use this to construct
// a visual render appropriate for the current platform.  A pointer
// to the presenter, and a platform specific parameter are given.
//

VisualRenderer* VisualRenderer::Create(Presenter* pPresenter,
									   void* pInfo)
{
	VisualRenderer* pVisRend = NULL;

// Construct a Windows 95/NT specific visual renderer
#ifdef WIN32
	pVisRend = new VisualRendererDD(pPresenter, (HWND)pInfo);
#endif

// Construct an SGI specific visual renderer
#if sgi
	pVisRend = new VisualRendererX11(pPresenter, (const char*)pInfo);
#endif

// Construct a sun specific visual renderer
#if sun
	pVisRend = new VisualRendererX11(pPresenter, (const char*)pInfo);
#endif

	pVisRend->m_pPresenter = pPresenter;

	return pVisRend;
} /* VisualRenderer::Create */


// ======================================================================

//
// VisualRenderer::VisualRenderer
//
// Constructor - empty the list of drawable nodes.
//

VisualRenderer::VisualRenderer()
{
	TRACE_PRINT("Entering VisualRenderer::VisualRenderer()\n");

	// Start with an empty array of drawable nodes
	m_drawableArray = NULL;
	m_drawableArraySize = 0;
	m_numberOfDrawables = 0;

	m_clearOnEnd = TRUE;
	m_fullRedrawRequired = FALSE;

	TRACE_PRINT("Leaving VisualRenderer::VisualRenderer()\n");
} /* VisualRenderer::VisualRenderer */


// ======================================================================

//
// VisualRenderer::~VisualRenderer
//
// Destructor - delete any memory associated with the list of drawable
// nodes.
//

VisualRenderer::~VisualRenderer()
{
	TRACE_PRINT("Entering VisualRenderer::~VisualRenderer()\n");

	if (m_drawableArray != NULL)
		delete (m_drawableArray);

	TRACE_PRINT("Leaving VisualRenderer::~VisualRenderer()\n");
} /* VisualRenderer::~VisualRenderer */


// ======================================================================

//
// VisualRenderer::Init
//
// Empty the list of drawable nodes (without deleting memory).
//

void VisualRenderer::Init()
{
	TRACE_PRINT("Entering VisualRenderer::Init()\n");

	m_numberOfDrawables = 0;

	TRACE_PRINT("Leaving VisualRenderer::Init()\n");
} /* VisualRenderer::Init */


// ======================================================================

//
// VisualRenderer::Terminate
//
// Empty the list of drawable nodes (without deleting memory).
//

void VisualRenderer::Terminate()
{
	TRACE_PRINT("Entering VisualRenderer::Terminate()\n");

	m_numberOfDrawables = 0;

	TRACE_PRINT("Leaving VisualRenderer::Terminate()\n");
} /* VisualRenderer::Terminate */


// ======================================================================

//
// VisualRenderer::AddToDrawList
//
// This function is called by a renderable node which needs to
// be rendered.
//

void VisualRenderer::AddToDrawList(DrawableNodePtr pDrawableNode)
{
	// If the array of drawable nodes is not large enough, make it larger
	if (m_numberOfDrawables >= m_drawableArraySize)
	{
		m_drawableArraySize += 10;
		m_drawableArray = (DrawableNodePtr*)realloc(m_drawableArray,
			m_drawableArraySize * sizeof(DrawableNodePtr));
	}

	// Add the details of this presentable media object to the list
	m_drawableArray[m_numberOfDrawables] = pDrawableNode;
	m_numberOfDrawables++;
} /* VisualRenderer:: AddToDrawList */


// ========================================================================

//
// CompareDrawables
//
// Compare the two given drawables, returning -1 if the first node
// is to be drawn before the second, and +1 if the second node is
// to be drawn before the first.
//
// The nodes are initially compared on their drawing order value.
// If this is the same, then the node which appeared first in the
// BIFS description is drawn first.
//

static int CompareDrawables(const void* elem1, const void* elem2)
{
	DrawableNodePtr drawable1 = *((DrawableNodePtr*)elem1);
	DrawableNodePtr drawable2 = *((DrawableNodePtr*)elem2);

	// Compare drawing orders
	if (drawable1->GetDrawingOrder() < drawable2->GetDrawingOrder())
		return -1;
	else if (drawable1->GetDrawingOrder() > drawable2->GetDrawingOrder())
		return +1;
	else
	{
		// If the drawing order is the same, then the object which
		// appeared first in the BIFS is drawn first.
		if (drawable1->GetCreateOrder() < drawable2->GetCreateOrder())
			return -1;
		else
			return +1;
	}
} /* CompareDrawables */


// ======================================================================

//
// VisualRenderer::DrawScene
//
// Draw all of the nodes which are present in the drawable list.
// Once the nodes have been drawn, this list is emptied.
//

void VisualRenderer::DrawScene()
{
	// Only proceed if there are drawables in the list
	if (m_numberOfDrawables == 0)
		return;

	TRACE_PRINT1("VR: Entering DrawScene(): %d drawables.\n",
				 m_numberOfDrawables);

	// Sort the list of drawables into draw order
	if (m_numberOfDrawables > 1)
		qsort(m_drawableArray, m_numberOfDrawables,
			  sizeof(DrawableNodePtr), CompareDrawables);

	// Start with two empty bounding rectangles (in BIFS coordinates)
	BRect needsDrawingRect;
	BRect transparentRect;

	// Go through the list of nodes, determine what area has changed,
	// and what area is transparent.
	unsigned int i;
	for (i = 0; i < m_numberOfDrawables; i++)
	{
		// If changed, add area of this node to the changed area
		if (m_drawableArray[i]->NeedsDrawing())
		{
#ifdef WRITE_TRACE
			TRACE_PRINT1("Node %d needs drawing, rect = ", i);
			m_drawableArray[i]->GetCombinedBoundingRect()->Print(traceFile);
			TRACE_PRINT("\n");
#endif
			needsDrawingRect.Union(m_drawableArray[i]->GetCombinedBoundingRect());
		}

		// If transparent, add area of this node to the transparent area
		if (m_drawableArray[i]->IsTransparent())
			transparentRect.Union(m_drawableArray[i]->GetCombinedBoundingRect());
	}

	// Ensure that these two rectangles are not larger than the screen
	// itself
	BRect screenBRect((int)GetScreenWidth() / -2, (int)GetScreenHeight() / 2,
					  (int)GetScreenWidth(), (int)GetScreenHeight());
	needsDrawingRect.Intersection(screenBRect);
	transparentRect.Intersection(screenBRect);

	// Reduce the transparent region to be only that area which is
	// both transparent and needs drawing.
	transparentRect.Intersection(needsDrawingRect);

/**/
#ifdef WRITE_TRACE
	TRACE_PRINT("VR: Needs drawing rect: ");
	needsDrawingRect.Print(traceFile);
	TRACE_PRINT("\n");
	TRACE_PRINT("VR: Transparent rect: ");
	transparentRect.Print(traceFile);
	TRACE_PRINT("\n");
#endif
/**/

	// Keep track of the region which is being drawn (used to
	// determine which nodes which have not changed still need to
	// be drawn).
	BRect drawnRect;
	
	// Only draw things if the changed rectangle is not empty
	if (!needsDrawingRect.IsEmpty())
	{
		// Erase the part of the screen which will have transparent
		// objects drawn in it.
		if (!transparentRect.IsEmpty())
		{
			TRACE_PRINT1("VR: Erasing rectangle: %s\n",
						 (char*)transparentRect);
			Clear(&transparentRect);
		}

		// And now draw the nodes, from the back to the front, only
		// drawing the part of the objects which lie within the changed
		// area.  Nodes which have not changed are only drawn if they
		// intersect with the transparent area, or something behind them
		// has been redrawn.
		drawnRect.MakeEmpty();
		for (i = 0; i < m_numberOfDrawables; i++)
		{
			const BRect* pBRect = m_drawableArray[i]->GetCombinedBoundingRect();

			if ((m_drawableArray[i]->NeedsDrawing()) ||
				(transparentRect.Overlaps(pBRect)) ||
				(drawnRect.Overlaps(pBRect)))
			{
				TRACE_PRINT1("VR: Drawing node %d.\n", i);
				m_drawableArray[i]->Draw(&needsDrawingRect);
				drawnRect.Union(pBRect);
			}
		}

		// And now copy the offscreen drawing area to the on screen,
		// only copying the area that has actually changed.
		TRACE_PRINT1("VR: Updating rectangle: %s\n",
					 (char*)needsDrawingRect);
		Paint(&needsDrawingRect);
	}

	TRACE_PRINT("VR: Leaving DrawScene()\n");
} /* VisualRenderer::DrawScene */


// ======================================================================

//
// VisualRenderer::FindNode
//
// Go through the list of drawable nodes, in front to back depth
// order (opposite to the drawing order), looking for the first which
// is under the given point (which is in BIFS coordinates).
//
// Currently the bounding rectangle is used to determine if the
// point is above the node.  This needs to be improved.
//

DrawableNode* VisualRenderer::FindNode(int x, int y)
{
	// Look through the list, starting with the front most
	for (int i = m_numberOfDrawables - 1; i >= 0; i--)
		if (m_drawableArray[i]->IsPointOver(x, y))
			return m_drawableArray[i];

	return NULL;
} /* VisualRenderer::FindNode */


// ======================================================================

// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
