/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Executive
FILE:       Executiv.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Impelentation of Executive
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "Im1defs.h"
#include "Executiv.h"
#include "Stream.h"
#include "AL.h"
#include "Service.h"
#include "File.h"
#include "Remote.h"
#include "BifsDec.h"
#include "BasePresenter.h"
#include "StreamCo.h"
#include "App.h"
#include "VisRend.h"
#include "Clock.h"
#include "Channel.h"

/*-----------------------------------------------------------------
Executive constructor. Saves the arguments, insantiate linked objects,
and create the Executive thread.
-----------------------------------------------------------------*/
Executive :: Executive (Application *pApp, HWND hWnd) :
	m_smpWaitToStart (1, 0),
	m_smpSceneBuilt (1, 0)
{
	m_bDelete = m_bActive = FALSE;
	m_pApp = pApp;
	m_hWnd = hWnd;
  	if (ZThread<Executive> (ExecutiveThread, this).Begin () == -1)
    	throw;
	m_pClock = new ClockReference;
	m_pBifs = new MediaStream (
				m_pClock,
				MPEG4_DEFAULT_BIFS_BUFFER_SIZE + 50
			  );
	m_pAL = new ALManager (this);
	m_pBifsDecoder = new BIFSDecoder (this, m_pBifs);
	m_pPresenter = Presenter::Create (this, hWnd);
	m_pScene = NULL;
	m_pService = NULL;

	CUSTOMIZE_EXECUTIVE_CONSTRUCTOR
}

/*-----------------------------------------------------------------
Executive destructor. Wait till the thread finishes then delete
all linked objects.
-----------------------------------------------------------------*/
Executive :: ~Executive ()
{
	CUSTOMIZE_EXECUTIVE_DESTRUCTOR

	m_bDelete = TRUE;
	m_pBifs -> Close();
	m_smpExecutiveActive.P ();
	delete m_pAL;
	delete m_pBifsDecoder;
	delete m_pBifs;
	delete m_pPresenter;
	delete m_pClock;
	delete m_pService;
	m_smpWaitToStart.V ();
	m_smpExecutiveThread.P ();
}

/*-----------------------------------------------------------------
Start() instantiates a service and start a connection.
-----------------------------------------------------------------*/
void Executive :: Start (SERVICE_TYPE type, LPCSTR szAddress)
{
	CUSTOMIZE_EXECUTIVE_START

	switch (type) {
	case FILE_SERVICE:
		m_pService = new FileService (m_pAL);
		break;
	case REMOTE_SERVICE:
		m_pService = new RemoteService (m_pAL);
		break;
	default:
		assert (0);
	}
	m_pService -> Connect (szAddress);
}

/*-----------------------------------------------------------------
Called by ALManager when connection done or failed. If done kicks off
the Executive thread.
-----------------------------------------------------------------*/
void Executive :: OnConnect (CONNECTION_RESULT result)
{
	if (result == CONNECTION_SUCCESS)
	{	m_nOpenStreams = 0;
		m_pClock -> Reset ();
		m_bActive = TRUE;
		m_smpExecutiveActive.P ();
		m_smpWaitToStart.V ();
	}
	else
	{	delete m_pService;
		m_pService = NULL;
	}
	m_pApp -> OnConnect (result);
}

/*-----------------------------------------------------------------
Stop() is called when player is closed by the user. Close the service.
Abort streams playing.
Wait till executive is not active, and shut down everything.
-----------------------------------------------------------------*/
void Executive :: Stop ()
{
	if (m_nOpenStreams > 0)
		m_smpOpenStreams.V ();
	if (m_pService)
		m_pService -> Close ();
	m_smpExecutiveActive.P ();
	if (m_bActive)
	{	m_bActive = FALSE;
		m_pPresenter -> Terminate ();
		delete m_pScene;
		m_pScene = NULL;	//ML added this line
	}
	m_smpExecutiveActive.V ();
}

/*-----------------------------------------------------------------
Finish is called from ALManager at end of session or error.
Wait till all streams are played to end, and notify the application.
-----------------------------------------------------------------*/
void Executive :: OnFinish (SERVICE_TERMINATION reason)
{
	//ML: Don't need the next two lines, as this is also done
	//    in ~Executive()
//	delete m_pService;
//	m_pService = NULL;
	m_smpOpenStreams.P ();
	m_smpOpenStreams.V ();
	m_pApp -> OnFinish (reason);
}

/*-----------------------------------------------------------------
Passes a paint requestion to the visual renderer
-----------------------------------------------------------------*/
void Executive :: Paint (void *info)
{
	m_pPresenter -> GetVisualRenderer () -> Paint (info);
}


/*-----------------------------------------------------------------
ExecutiveThread() is the thread function of Executive
-----------------------------------------------------------------*/
void Executive :: ExecutiveThread ()
{
	/*-----------------------------------------------------------------
	Mark the thread as running, then loop till object destructor performs
	-----------------------------------------------------------------*/
	m_smpExecutiveThread.P ();
	for (;;)
	{	
		/*-----------------------------------------------------------------
		Wait for Start command, then open the BIFS stream.
		-----------------------------------------------------------------*/
		m_smpWaitToStart.P ();
		if (m_bDelete)
			break;
		m_registeredNodes.Clear ();
		m_pBifs -> Reset ();
		DataChannel *pChannel =
			m_pAL -> AttachStream (1 << 6, m_pBifs, NULL);
		if (pChannel)
		{	m_pAL -> RequestChannels ();
			pChannel -> ChannelReady ();
		}

		/*-----------------------------------------------------------------
		Build a scene, Initialize the Presenter, then update the scene
		-----------------------------------------------------------------*/
		m_pScene = m_pBifsDecoder -> BuildScene ();
		m_smpSceneBuilt.V ();
		if (m_pScene)
		{	m_pAL -> RequestChannels ();
			m_pPresenter -> Init (m_pScene);
			m_pBifsDecoder -> UpdateScene ();
			m_smpExecutiveActive.V ();
		}
		else
		{
			m_smpExecutiveActive.V ();
			Stop ();
		}
	}
	m_smpExecutiveThread.V ();
}

/*-----------------------------------------------------------------
LockScene() locks the scene. If bWait is true, it waits till the scene
is released by other threads. If it's false, it returns FALSE if 
scene is currently resereved.
-----------------------------------------------------------------*/
BOOL Executive :: LockScene (BOOL bWait)
{
	if (!bWait)
		return m_mtxScene.Try ();
	else
	{	m_mtxScene.P ();
		return TRUE;
	}
}

/*-----------------------------------------------------------------
Manage stream count which is used to make sure session is not finished
before streams are played to the end
-----------------------------------------------------------------*/
void Executive :: StreamOpen ()
{
	if (m_nOpenStreams++ == 0)
		m_smpOpenStreams.P ();
}

void Executive :: StreamClose ()
{
	if (--m_nOpenStreams == 0)
		m_smpOpenStreams.V ();
}

/*-----------------------------------------------------------------
A node of type StreamConsumer registers here when it is instantiated
-----------------------------------------------------------------*/
void Executive :: RegisterStreamConsumer (StreamConsumer *pNode)
{
	/*-----------------------------------------------------------------
	Add to the list of registered nodes and increment the number of
	nodes which have not been attached to streams by 1.
	-----------------------------------------------------------------*/
	m_registeredNodes += pNode;
}

/*-----------------------------------------------------------------
A node of type StreamConsumer unregisters here when it is destroyed
-----------------------------------------------------------------*/
void Executive :: UnregisterStreamConsumer  (StreamConsumer *pNode)
{
	m_registeredNodes -= pNode;
}

/*-----------------------------------------------------------------
BIFSDecoder calls this function when an OD is parsed. Find the 
corresponding node in the list of registered nodes and attach the
stream to it.
-----------------------------------------------------------------*/
void Executive :: SetStream (
	int nObjectDescriptorId,
	ES_DESCRIPTOR *pES_Descriptor,
	LPCSTR szFileName
)
{
	LockScene ();
	for (int i = m_registeredNodes.GetLength (); --i >= 0;)
	{
		if (m_registeredNodes [i] -> GetObjectDescriptorID () ==
			nObjectDescriptorId
		)
		{	
			m_registeredNodes [i] ->
				SetStream (pES_Descriptor, szFileName);
			break;
		}
	}
	UnlockScene ();
}

/*-----------------------------------------------------------------
Pass user events to the presenter
-----------------------------------------------------------------*/
void Executive :: OnUserEvent  (
	USER_EVENT event,
	int x,
	int y,
	UINT uControlKeysStatus
)
{
	m_pPresenter -> OnUserEvent (event, x, y, uControlKeysStatus);
}