/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  DMIF
FILE:       File.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of FileService
-------------------------------------------------------------------------*/
#include "defs.h"
#include "File.h"
#include "AL.h"
#include "Except.h"

/*-----------------------------------------------------------------
Constructor
-----------------------------------------------------------------*/
FileService :: FileService (ALManager *pAL) :
	Service (pAL),
	m_smpWaitForChannel (1, 0)
{
	m_nChannelWaitedFor = -1;
}

/*-----------------------------------------------------------------
Destructor
-----------------------------------------------------------------*/
FileService :: ~FileService ()
{
	//ML added these lines to ensure that the thread has finished
	m_smpThread.P();
	m_smpThread.V();
}

/*-----------------------------------------------------------------
Open the file, load the association table, start the file thread,
and notify the application
-----------------------------------------------------------------*/
void FileService :: Connect (LPCSTR szFileName)
{
	if ((m_fInput = fopen (szFileName, "rb")) == NULL)
	{	GetALManager () -> OnConnect (CANNOT_OPEN_FILE);
		return;
	}

	try
	{
		int nTableLength = GetByte ();

		while (nTableLength-- > 0)
		{	
			GetByte ();
			int c = GetByte ();
			int nESid = MAKEWORD (GetByte (), c);
			m_channelMap.ElementAt (GetByte ()).m_nESid = nESid;
		}
	}
	catch (DemuxException)
	{
		fclose (m_fInput);
		GetALManager () -> OnConnect (UNEXPECTED_EOF);
		return;
	}

	m_bActive = TRUE;
  	if (ZThread<FileService> (
			FileThread,
			this,
			THREAD_PRIORITY_BELOW_NORMAL
		).Begin () == -1
	)
	{	fclose (m_fInput);
		GetALManager () -> OnConnect (MEMORY_PROBLEM);
		return;
	}

	GetALManager () -> OnConnect (CONNECTION_SUCCESS);
}

/*-----------------------------------------------------------------
Signal the file thread to terminate
-----------------------------------------------------------------*/
void FileService :: Close ()
{
	m_bActive = FALSE;
	if (m_nChannelWaitedFor >= 0)
	{	m_nChannelWaitedFor = -1;
		m_smpWaitForChannel.V ();
	}
}

/*-----------------------------------------------------------------
When channels are requested, the function calls Service::OnChannelOpen()
for each requested channel.
-----------------------------------------------------------------*/
void FileService :: OpenChannels (
	int nNumberOfChannels,
	const int *pESidList
)
{
	SMT smt;
	StreamMapElement &rMapElement = smt.ElementAt (0);

	while (nNumberOfChannels-- > 0)
	{	
		int nESid = *pESidList++;
		int i = m_channelMap.Find (ChannelAssociation (nESid));
		rMapElement.m_nAssociationTag = i;
		rMapElement.m_nFmcNo = -1;
		rMapElement.m_nESid = nESid;
		OnChannelOpen ((HANDLE) i, smt);
	}
}

/*-----------------------------------------------------------------
Mark each of the channels as "ready"
-----------------------------------------------------------------*/
void FileService :: ChannelsReady (
	int nNumberOfChannels,
	const int *pAssociationTagList
)
{
	while (nNumberOfChannels-- > 0)
	{	
		int nTag = *pAssociationTagList++;
		if (nTag >= 0)
		{	m_mtxChannels.P ();
			m_channelMap [nTag].m_bReady = TRUE;

			/*-----------------------------------------------------------------
			If the file thread is waiting for this channel to be ready,
			signal it to go
			-----------------------------------------------------------------*/
			if (m_nChannelWaitedFor == nTag)
			{	m_smpWaitForChannel.V ();
				m_nChannelWaitedFor = -1;
			}
			m_mtxChannels.V ();
		}
	}	
}

/*-----------------------------------------------------------------
The file thread reads data from the file and pass it to Service
-----------------------------------------------------------------*/
void FileService :: FileThread ()
{
	BYTE buffer [256];

	// Mark the thread as running
	m_smpThread.P();

	while (m_bActive)
	{
		try
		{	
			/*-----------------------------------------------------------------
			If stream transmission has not been requested yet, wait till it
			is requested
			-----------------------------------------------------------------*/
			int nFmcNo = GetByte ();
			m_mtxChannels.P ();		// critical section
			if (!m_channelMap [nFmcNo].m_bReady)
			{	
				m_nChannelWaitedFor = nFmcNo;
				m_mtxChannels.V ();
				m_smpWaitForChannel.P ();
				if (!m_bActive)
					break;
			}
			else
				m_mtxChannels.V ();

			UINT nLength = GetByte ();
			if (fread (buffer, 1, nLength, m_fInput) < nLength)
				throw DemuxException ();
			OnDataReceived ((HANDLE) nFmcNo, buffer, nLength);
		}
		catch (DemuxException)
		{
			m_bActive = FALSE;
		}
	}
	GetALManager () -> OnFinish (END_OF_SERVICE);

	// Flag that the thread is terminating
	m_smpThread.V();
}

/*-----------------------------------------------------------------
Read one byte and throw exception on end of file
-----------------------------------------------------------------*/
int FileService :: GetByte ()
{
	int c = fgetc (m_fInput);
	if (c == EOF)
		throw DemuxException ();
	return c;
}