/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor / Presenter
FILE:       MediaObj.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of base class for MediaObject objects.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "MediaObj.h"
#include "Executiv.h"
#include "BifsDec.h"
#include "Route.h"
#include "BasePresenter.h"
#include "NodeFld.h"
#include "bifs.h"
#include "Group.h"
#include "Root.h"
#include "Quantize.h"

/*-----------------------------------------------------------------
MediaObject constructor
-----------------------------------------------------------------*/
MediaObject :: MediaObject ()
{
	m_pRoot = NULL;
	m_nUse = 0;
	m_pFirstRoute = NULL;
	m_nID = -1;
	m_pExecutive = NULL;
	m_bIsRoot = FALSE;

	CUSTOMIZE_MEDIAOBJECT_CONSTRUCTOR
}

/*-----------------------------------------------------------------
MediaObject destructor. First flush out everything from the stream
(to make sure no object is waiting on the stream, then delete the
associated decoder and stream.
-----------------------------------------------------------------*/
MediaObject :: ~MediaObject ()
{
	CUSTOMIZE_MEDIAOBJECT_DESTRUCTOR

	/*-----------------------------------------------------------------
	Delete all routes which this is their arrival node
	-----------------------------------------------------------------*/
	while (m_pFirstRoute)
		delete m_pFirstRoute;  // this unlinks the route automatically

	/*-----------------------------------------------------------------
	If this node has an nID unregister it from the root
	-----------------------------------------------------------------*/
	if (m_pRoot && m_nID >= 0)
		m_pRoot -> UnregisterNode (this);

	/*-----------------------------------------------------------------
	Acknowledge all containers that this node does not exist
	-----------------------------------------------------------------*/
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
			if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode (this);
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode ();
}

/*-----------------------------------------------------------------
Store the Executive pointer and perform object initialization
-----------------------------------------------------------------*/
void MediaObject :: SetExecutive (Executive *pExecutive)
{
	m_pExecutive = pExecutive;

	CUSTOMIZE_MEDIAOBJECT_INITIALIZATION
}

/*-----------------------------------------------------------------
SetRoot stores a pointer to the root object of this object.
If this object has an ID, it registers it with the root.
-----------------------------------------------------------------*/
void MediaObject :: SetRoot (Root *pRoot)
{
	m_pRoot = pRoot;
	if (m_nID >= 0 && pRoot)
		pRoot -> RegisterNode (this);
}

/*-----------------------------------------------------------------
Set the node ID and register it with the node if necessary
-----------------------------------------------------------------*/
void MediaObject :: SetID (int nID)
{
	m_nID = nID;
	if (nID >= 0 && m_pRoot)
		m_pRoot -> RegisterNode (this);
}		

/*-----------------------------------------------------------------
Parse a MediaObject by using the field table.
-----------------------------------------------------------------*/
BOOL MediaObject :: Parse (BIFSDecoder *pBifs)
{
	TRACEF ("%s {", GetName());
	TRACENEWLINE;
	TRACEINDENT (1);

	int nFieldCount = GetBifsBounds () -> m_nTotalDefs;

	/*-----------------------------------------------------------------
	A node may be described by a mask description or by a list 
	description
	-----------------------------------------------------------------*/
	if (pBifs -> ParseInt (1))  // if MaskAccess
	{	ZArray<BYTE> mask;
		for (int i = 0; i < nFieldCount; i++)
			mask.SetAt (i, pBifs -> ParseInt (1));
		
		/*-----------------------------------------------------------------
		Now loop for each field.
		-----------------------------------------------------------------*/
		for (i = 0; i < nFieldCount; i++)
		{
			if (mask [i] && !ParseField (pBifs, GetBifsDefEntry (i)))
				return FALSE;
		}
	}
	else		// ListNodeDescription
	{
		/*-----------------------------------------------------------------
		Now loop for each field.
		-----------------------------------------------------------------*/
		while (!pBifs -> ParseInt (1))	// while !endFlag
		{
			int nIndex =
				pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerDef);
			if (!ParseField (pBifs, GetBifsDefEntry (nIndex)))
				return FALSE;
		}
	}

	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
	SetUpdateTime ();	
	return TRUE;
}

/*-----------------------------------------------------------------
Add a Route to the node (this is the arrival node)
-----------------------------------------------------------------*/
void MediaObject :: LinkRoute (Route *pRoute)
{
	pRoute -> LinkArrival (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Unlink a Route from the list of routes (this is the arrival node)
-----------------------------------------------------------------*/
void MediaObject :: UnlinkRoute (Route *pRoute)
{
	pRoute -> UnlinkArrival (m_pFirstRoute);
}

/*-----------------------------------------------------------------
A route to this node was activated. If this is a field, update the
field and notify the node. If this is an eventIn, activate the event
handling function.
-----------------------------------------------------------------*/
void MediaObject :: RouteEvent (NodeField *pField, int nFieldIndex)
{
	if (nFieldIndex < 0 || nFieldIndex >= GetFieldCount ())
		throw BifsException (BIFX_PARSING_ERROR);
	*GetField (nFieldIndex) = *pField;
	UpdateField (GetFieldTable () + nFieldIndex);
}

/*-----------------------------------------------------------------
Calls Presenter :: GetTime () to retrieve the renderer "time", and
stamp the node with this value.  Used by the renderer to identify new
scene updates.
-----------------------------------------------------------------*/
inline void MediaObject :: SetUpdateTime ()
{
	m_dwUpdateTime = m_pExecutive -> GetPresenter () -> GetTime ();
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to insert a value into one of the MFxxx fields
of the node.  The field ID, the new value and the exact insertion
position are parsed from the BIFS stream by this function. Returns TRUE
if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: InsertValue (BIFSDecoder *pBifs)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerIn);
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	Quantizer quant (
		pEntry -> m_fMin,  
		pEntry -> m_fMax,  
		pEntry -> m_nDimension, 
		pEntry -> m_nQuantizationType
	);
	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pBifs -> ParseInt (2)) {
	case 0:				// insertion at a specified position
		nPos = pBifs -> ParseInt (16);
		TRACEF ("[%d] ", nPos);
		bResult = pField -> InsertElement (pBifs, this, nPos, &quant);
		break;
	case 1: 			// insertion at the beginning of the field
		TRACEF ("[0] ", 0);
		bResult = pField -> InsertElement (pBifs, this, 0, &quant);
		break;
	case 2: 			// insertion at the end of the field
		TRACEF ("[end] ", 0);
		bResult = pField -> AppendElement (pBifs, this, &quant);
		break;
	}
	if (bResult)
	{	UpdateField (pEntry);
		pField -> OnFieldUpdate ();
	}
	return bResult;
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to delete a value from one of the MFxxx fields 
of the node.  The field ID and the element index are parsed from the
BIFS by this function.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: DeleteValue (BIFSDecoder *pBifs)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerIn);
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pBifs -> ParseInt (2)) {
	case 0:				// deletion from a specified position
		nPos = pBifs -> ParseInt (16);
		TRACEF ("[%d] ", nPos);
		if (nPos < pField -> GetLength ())
		{	pField -> DeleteFrom (nPos);
			bResult = TRUE;
		}
		break;
	case 1: 			// deletion from the beginning of the field
		TRACEF ("[0] ", 0);
		if (pField -> GetLength () > 0)
		{	pField -> DeleteFrom (0);
			bResult = TRUE;
		}
		break;
	case 2: 			// delete last element of field
		TRACEF ("[end] ", 0);
		if (pField -> GetLength () > 0)
		{	pField -> DeleteFrom (pField -> GetLength () - 1);
			bResult = TRUE;
		}
		break;
	}
	return bResult;
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to replace a value of one of the nodes fields. 
The field ID and new value are parsed from the BIFS by this function.  
Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: ReplaceValue (BIFSDecoder *pBifs)
{
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerIn);
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	return ParseField (pBifs, pEntry);
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to replace a value of one of the elements of 
one of the MFxxx fields of the node.  The field ID, the element index 
and the new value are parsed from the BIFS by this function.  Returns 
TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: ReplaceIndexedValue (BIFSDecoder *pBifs)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerIn);
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	Quantizer quant (
		pEntry -> m_fMin,  
		pEntry -> m_fMax,  
		pEntry -> m_nDimension, 
		pEntry -> m_nQuantizationType
	);
	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pBifs -> ParseInt (2)) {
	case 0:				// replacement a specified position
		nPos = pBifs -> ParseInt (16);
		TRACEF ("[%d] ", nPos);
		bResult = pField -> ReplaceElement (pBifs, this, nPos, &quant);
		break;
	case 1: 			// replacement at the beginning of the field
		TRACEF ("[0] ", 0);
		bResult = pField -> ReplaceElement (pBifs, this, 0, &quant);
		break;
	case 2: 			// Replacement at the end of the field
		TRACEF ("[end] ", 0);
		assert (pField -> GetLength () > 0);
		bResult = pField -> ReplaceElement (
			pBifs,
			this,
			pField -> GetLength () - 1,
			&quant
		);
		break;
	}
	if (bResult)
	{	UpdateField (pEntry);
		pField -> OnFieldUpdate ();
	}
	return bResult;
}

/*-----------------------------------------------------------------
Add a new Route whose departure field is one of the fields of this 
node.  This departure field ID and arrival node ID are parsed from the 
BIFS stream, then BindNode() is called in order to link the new Route 
to the proper arrival field.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: AddRoute (int nRouteID,  BIFSDecoder *pBifs)
{
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerOut);
	const BifsFieldTable *pEntry = GetBifsOutEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	TRACEF ("%s TO ", pEntry -> m_szName);
	Route *pRoute = new Route (nRouteID, GetField (pEntry));

	int nNodeID = pBifs -> ParseInt (10);
	TRACEF ("%d.", nNodeID);
	MediaObject *pNode = GetRoot () -> FindNode (nNodeID);
	if (pNode == NULL)
		return FALSE;
	return pNode -> BindRoute (pRoute, pBifs);
}

/*-----------------------------------------------------------------
Having instantiated a new Route, and binding it to a departure field, 
AddRoute() calls this function where it requests the node to bind the 
Route to one of its fields as an arrival field.  The arrival field index 
is parsed from the BIFS stream.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: BindRoute (Route *pRoute, BIFSDecoder *pBifs)
{
	int nIndex =
		pBifs -> ParseInt (GetBifsBounds () -> m_nBitsPerIn);
	const BifsFieldTable *pTable = GetFieldTable ();

	for (int i = GetFieldCount (); --i >= 0;)
		if (nIndex == pTable [i].m_inID)
		{	TRACEF ("%s", pTable [i].m_szName);
			pRoute -> Bind (this, i);
			return TRUE;
		}

	return FALSE;
}

/*-----------------------------------------------------------------
This function is called after recalculation of the bounding-box.  The 
function calls UpdateBBox() on all the nodes parents.
-----------------------------------------------------------------*/
void MediaObject :: OnBBoxUpdate ()
{
	for (int i = m_parents.GetLength (); --i >= 0;)
		m_parents [i] -> UpdateBBox ();
}

/*-----------------------------------------------------------------
Add the container field to the list of containers and the parent node
to the list of parents
-----------------------------------------------------------------*/
void MediaObject :: SetContainer (
	NodeField *pContainer,
	BOOL bISMF,
	GroupNode *pParent
)
{
	m_containers += Container (pContainer, bISMF);
	SetParent (pParent);
}

/*-----------------------------------------------------------------
Remove the container field from the list of containers. If the node
is not contained any more in any other node, delete it.
-----------------------------------------------------------------*/
void MediaObject :: RemoveContainer (NodeField 
	*pContainer,
	GroupNode *pParent
)
{
	RemoveParent (pParent);
	m_containers -= pContainer;
	if (m_containers.GetLength () == 0)
		delete this;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose defID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsDefEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_defID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose inID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsInEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_inID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose outID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsOutEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_outID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose dynID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsDynEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_dynID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Parse a node field, when a pointer to the BIFS table entry describing
that field is given
-----------------------------------------------------------------*/
BOOL MediaObject :: ParseField (
	BIFSDecoder *pBifs,
	const BifsFieldTable *pEntry
)
{	
	if (pEntry == NULL)
		return FALSE;
	Quantizer quant (
		pEntry -> m_fMin,  
		pEntry -> m_fMax,  
		pEntry -> m_nDimension, 
		pEntry -> m_nQuantizationType
	);

	TRACEF ("%s ", pEntry -> m_szName);
	BOOL bResult = GetField (pEntry) -> Parse (pBifs, this, &quant);
	TRACENEWLINE;
	return bResult;
}

/*-----------------------------------------------------------------
A node is replaced by replaceNode BIFS command.
-----------------------------------------------------------------*/
void MediaObject :: ReplaceNode (BIFSDecoder *pBifs)
{
	/*-----------------------------------------------------------------
	Parse a new node, then go through all the containers and replace
	the pointer to the old node by the new one
	-----------------------------------------------------------------*/
	SFGenericNode node;
	node.Parse (pBifs, this);
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
		{	if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					ReplaceNode (this, node);
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					ReplaceNode (node);
			SetContainer (m_containers [i].m_pContainer);
			m_containers [i].m_pContainer = NULL;
		}

	/*-----------------------------------------------------------------
	Copy the list of parents to the new node
	-----------------------------------------------------------------*/
	for (i = m_parents.GetLength (); --i >= 0;)
		node -> SetParent (m_parents [i]);

	/*-----------------------------------------------------------------
	Call OnUpdate on the new node then delete this (the old) node
	-----------------------------------------------------------------*/
	node -> SetUpdateTime ();
	node -> OnUpdate ();
	delete this;
}

/*-----------------------------------------------------------------
A node was deleted by a DeleteNode BIFS command
-----------------------------------------------------------------*/
void MediaObject :: Delete ()
{
	/*-----------------------------------------------------------------
	Notify all the containers that the node was deleted
	-----------------------------------------------------------------*/
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
		{	if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode (this);
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode ();
			m_containers [i].m_pContainer -> OnFieldUpdate ();
			m_containers [i].m_pContainer = NULL;
		}

	/*-----------------------------------------------------------------
	Notify all parents that node was deleted
	-----------------------------------------------------------------*/
	for (i = m_parents.GetLength (); --i >= 0;)
	{	m_parents [i] -> SetUpdateTime ();
		m_parents [i] -> OnUpdate ();
	}

	/*-----------------------------------------------------------------
	Delete the node
	-----------------------------------------------------------------*/
	delete this;
}

/*-----------------------------------------------------------------
A field was updated as a result of a Route or BIFS update
pEntry points to the entry in the BIFS table describing the field
-----------------------------------------------------------------*/
void MediaObject :: UpdateField (const BifsFieldTable *pEntry)
{
	EventHandler function = pEntry -> m_function;
	if (function == NULL)
	{	SetUpdateTime ();
		OnUpdate ();
	}
	else
		(*this.*function) ();
}