#if !defined(_MediaObj_h)	// Sentry, use file only if it's not already included.
#define _MediaObj_h

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor / Presenter
FILE:       MediaObj.h
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Declarations of base classes for MediaObject objects.
-------------------------------------------------------------------------*/
#include "Im1defs.h"
#include "ztl.h"
#include "custom.h"

class Executive;
class BIFSDecoder;
class Route;
class NodeField;
class GroupNode;
class Root;
struct BifsFieldTable;
struct BifsBounds;

#ifdef BIFSENC
class BIFSEncoder;
#endif

/*-------------------------------------------------------------------------
- MediaObject is the base class for constructing scene nodes.
- MediaObject objects can parse themselves from BIFS.
- MediaObject objects can render themselves.
- Each object should have default constructor, because during parsing,
the object is created first, then it loads itself.
-------------------------------------------------------------------------*/

class MediaObject
{
	int m_nID;
	Executive *m_pExecutive;
	Root *m_pRoot;
	Route *m_pFirstRoute;
	DWORD m_dwUpdateTime;
	int m_nUse;
	ZArray<GroupNode *> m_parents;

	struct Container
	{	NodeField *m_pContainer;
		BOOL m_bISMF;
		Container (NodeField *pContainer = NULL, BOOL bISMF = FALSE) :
			m_pContainer (pContainer),
			m_bISMF (bISMF)
		{}
		BOOL operator == (NodeField *pContainer) const
		{
			return pContainer == m_pContainer;
		}
		operator NodeField * () const {return m_pContainer;}
	};
	ZArray<Container> m_containers;

private:
	const BifsFieldTable * GetBifsDefEntry (int nID);
	const BifsFieldTable * GetBifsInEntry (int nID);
	const BifsFieldTable * GetBifsOutEntry (int nID);
	const BifsFieldTable * GetBifsDynEntry (int nID);
	virtual const BifsBounds *GetBifsBounds () const = 0;
	NodeField *GetField (const BifsFieldTable *pEntry);

protected:
	virtual int GetFieldCount () const = 0;
	virtual const BifsFieldTable *GetFieldTable () const = 0;
	NodeField *GetField (int nIndex);

#ifndef BIFSENC
private:
	BOOL ParseField (BIFSDecoder *pBifs, const BifsFieldTable *pEntry);

protected:
	BOOL m_bIsRoot;
	void UpdateField (const BifsFieldTable *pEntry);
	void OnBBoxUpdate ();

public:
	BOOL IsRoot () {return m_bIsRoot;}
	void Delete ();
	void ReplaceNode (BIFSDecoder *pBifs);
	Executive *GetExecutive () {return m_pExecutive;}
	BOOL BindRoute (Route *pRoute,  BIFSDecoder *pBifs);
	BOOL AddRoute (int nRouteID,  BIFSDecoder *pBifs);
	BOOL ReplaceIndexedValue (BIFSDecoder *pBifs);
	BOOL ReplaceValue (BIFSDecoder *pBifs);
	BOOL DeleteValue (BIFSDecoder *pBifs);
	BOOL InsertValue (BIFSDecoder *pBifs);
	MediaObject ();
	virtual ~MediaObject ();
	void SetExecutive (Executive *pExecutive);
	void SetRoot (Root *pRoot);
	void SetID (int nID);
	virtual BOOL Parse (BIFSDecoder *pBifs);
	virtual BOOL Render (void * = NULL) {return TRUE;}
	Root *GetRoot () {return m_pRoot;}
	int GetID () {return m_nID;}
	void LinkRoute (Route *pRoute);
	void UnlinkRoute (Route *pRoute);
	void RouteEvent (NodeField *pField, int nFieldIndex);
	void SetUpdateTime ();
	DWORD GetUpdateTime () {return m_dwUpdateTime;}
	virtual void OnUpdate () {}
	virtual BOOL Activate () {return FALSE;}
	void SetParent (GroupNode *pParent)
	{
		if (pParent)
			m_parents += pParent;
	}
	GroupNode *GetParent(int index = 0) const
	{
		return index < m_parents.GetLength () ? m_parents [index] : NULL;
	}
	void RemoveParent (GroupNode *pParent)
	{
		if (pParent)
			m_parents -= pParent;
	}
	void SetContainer (
		NodeField *pContainer,
		BOOL bISMF = FALSE,
		GroupNode *pParent = NULL
	);
	void RemoveContainer (NodeField *pContainer, GroupNode *pParent = NULL);
		
	CUSTOMIZE_MEDIAOBJECT_DECLARATION
#endif

#ifdef _DEBUG
public:
	virtual char *GetName () {return NULL;}
#endif


#ifdef BIFSENC
public:
	BOOL Encode (BIFSEncoder *pBifs);
	BOOL EncodeField (BIFSEncoder *pBifs, const BifsFieldTable *pEntry);
	void EncodeFieldElement (
		BIFSEncoder *pBifs,
		const BifsFieldTable *pEntry
	);
	void EncodeRoute (BIFSEncoder *pBifs, BOOL bIn);
	virtual int GetNodeCode () = 0;
	void InsertToField (BIFSEncoder *pBifs, BOOL bAppend);
	void IndexedValueDeletion (BIFSEncoder *pBifs, BOOL bLast);
	void FieldReplacement (BIFSEncoder *pBifs, BOOL bLast);
	int GetID () {return m_nID;}
	void SetID (int nID) {m_nID = nID;}

private:
	const BifsFieldTable *GetBifsEntry (LPCSTR szName);
	const BifsFieldTable *ParseMFElement (BOOL bLast, int &rIndex);
#endif
};

template <class T, int nodeType> class NodeFactory
{
	static int m_nodeType;

public:
	operator int () const {return m_nodeType;}
};

#endif                                      // _MediaObj_h sentry.
