/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       NodeFld.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of NodeField and derivations.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "NodeFld.h"
#include "MediaObj.h"
#include "BifsDec.h"
#include "Route.h"
#include "Root.h"
#include "Quantize.h"

/*-----------------------------------------------------------------
NodeField destructor. Delete all routes
-----------------------------------------------------------------*/
NodeField :: ~NodeField ()
{
	while (m_pFirstRoute)
		delete m_pFirstRoute;  // this unlinks the route automatically
}

/*-----------------------------------------------------------------
Add a Route to the field
-----------------------------------------------------------------*/
void NodeField :: LinkRoute (Route *pRoute)
{
	pRoute -> Link (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Unlink a Route from the list of routes
-----------------------------------------------------------------*/
void NodeField :: UnlinkRoute (Route *pRoute)
{
	pRoute -> Unlink (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Activate attached routes on field update
-----------------------------------------------------------------*/
void NodeField :: OnFieldUpdate ()
{
	for (Route *pRoute = m_pFirstRoute;
		pRoute != NULL;
		pRoute = pRoute -> GetNext ()
	)
		pRoute -> Activate ();
}

/*-----------------------------------------------------------------
Parse an SFInt
-----------------------------------------------------------------*/
BOOL SF<int> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject*, 
	Quantizer *pQuant
)
{
	/*-----------------------------------------------------------------
	Look at the Quantizer to see if min-max values applied
	-----------------------------------------------------------------*/
	m_value = pQuant -> ParseInt (pBifs);
	TRACEF ("%d", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFBool
-----------------------------------------------------------------*/
BOOL SF<Bool> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pBifs -> ParseInt (1);
	TRACEF ("%s", m_value ? "TRUE" : "FALSE");
	return TRUE;
}

/*-----------------------------------------------------------------
Parse and SFFloat
-----------------------------------------------------------------*/
BOOL SF<float> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pBifs -> ParseFloat();
	TRACEF ("%.2f", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFTime
-----------------------------------------------------------------*/
BOOL SF<Time> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pBifs -> ParseDouble();
	TRACEF ("%.2lf", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec3f
-----------------------------------------------------------------*/
BOOL SF<Vec3f> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.x = pBifs -> ParseFloat();
	m_value.y = pBifs -> ParseFloat();
	m_value.z = pBifs -> ParseFloat();
	TRACEF3 ("%.2f %.2f %.2f", m_value.x, m_value.y, m_value.z);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec2f
-----------------------------------------------------------------*/
BOOL SF<Vec2f> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.x = pBifs -> ParseFloat();
	m_value.y = pBifs -> ParseFloat();
	TRACEF2 ("%.2f %.2f", m_value.x, m_value.y);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation
-----------------------------------------------------------------*/
BOOL SF<Rotation> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.x = pBifs -> ParseFloat();
	m_value.y = pBifs -> ParseFloat();
	m_value.z = pBifs -> ParseFloat();
	m_value.theta = pBifs -> ParseFloat();
	TRACEF4 ("%.2f %.2f %.2f %.2f", m_value.x, m_value.y, m_value.z, m_value.theta);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation2D
-----------------------------------------------------------------*/
BOOL SF<Rotation2D> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.x = pBifs -> ParseFloat();
	m_value.y = pBifs -> ParseFloat();
	m_value.theta = pBifs -> ParseFloat();
	TRACEF3 ("%2.f %.2f %.2f", m_value.x, m_value.y, m_value.theta);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFColor
-----------------------------------------------------------------*/
BOOL SF<Color3f> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.r = pBifs -> ParseFloat();
	m_value.g = pBifs -> ParseFloat();
	m_value.b = pBifs -> ParseFloat();
	TRACEF3 ("%.2f %.2f %.2f", m_value.r, m_value.g, m_value.b);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFObjectID
-----------------------------------------------------------------*/
BOOL SFObjectID :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pBifs -> ParseInt (10);
	TRACEF ("%d", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse SFString
-----------------------------------------------------------------*/
BOOL SF<ZString> :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	//ML added this line to empty the string
	// (not a permanent solution as it will affect APPEND)
	m_value.Cut(0);

	int nLength = pBifs -> ParseInt (8);
	while (nLength--)
		m_value += pBifs -> ParseInt (8);
	TRACEF ("%s", LPCSTR (m_value));
	return TRUE;
}

/*-----------------------------------------------------------------
A dummy function
-----------------------------------------------------------------*/
BOOL SF<MediaObject *> :: Parse (
	BIFSDecoder *,
	MediaObject *, 
	Quantizer *
)
{
	return FALSE;
}

/*-----------------------------------------------------------------
SFGenericNode Destructor.  Deletes the MediaObject, unless it is defined
by other node and only USEd here.
-----------------------------------------------------------------*/
SFGenericNode :: ~SFGenericNode ()
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
}

/*-----------------------------------------------------------------
Assign a value into a generic-node. If the old value is not a copy 
of another field, delete the old node
-----------------------------------------------------------------*/
MediaObject *SFGenericNode :: operator = (MediaObject *node)
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
	m_pNode = node;
	if (node)
		node -> SetContainer (this);
	OnFieldUpdate ();
	return node;
}

/*-----------------------------------------------------------------
Parse a generic node
-----------------------------------------------------------------*/
BOOL SFGenericNode :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject *pNode, 
	Quantizer * /*pQuant*/
)
{
	if (pBifs -> ParseInt (1))	// if isReused
	{	int nID = pBifs -> ParseInt (10);
		TRACEF ("USE %d\n", nID);
		if (pNode &&
			(m_pNode = pNode -> GetRoot () -> FindNode (nID)) != NULL
		)
			m_pNode -> SetContainer (this);
		else
			return FALSE;
	}
	else
	{	
		/*-----------------------------------------------------------------
		Get object type, instantiate the object, and let it parse itself
		-----------------------------------------------------------------*/
		int nID = -1;
		int nNodeType = pBifs -> ParseInt (BitFieldWidth (GetNodeTableSize ()));
		if (pBifs -> ParseInt (1)) // if isUpdatable
		{	
			nID = pBifs -> ParseInt (10);
			TRACEF ("DEF %d ", nID);
		}
	
		m_pNode = CreateNode (nNodeType);
		if (!m_pNode)
			return FALSE;
		m_pNode -> SetExecutive (pBifs -> GetExecutive ());
		m_pNode -> SetID (nID);

		/*-----------------------------------------------------------------
		The root of the new object is the same as of the container 
		object. If the container object is itslef a root, the root is
		the container object.
		-----------------------------------------------------------------*/
		if (pNode)
			m_pNode -> SetRoot (pNode -> GetRoot ());
		else
		{	assert (m_pNode -> IsRoot ());
			m_pNode -> SetRoot ((Root *) m_pNode);
		}
		m_pNode -> SetContainer (this);
		m_pNode -> Parse (pBifs);
	}
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an MF field
-----------------------------------------------------------------*/
BOOL MField :: Parse (
	BIFSDecoder *pBifs, 
	MediaObject *pNode, 
	Quantizer *pQuant
)
{
	TRACEF ("[", 0);
	TRACENEWLINE;
	TRACEINDENT (1);

	/*-----------------------------------------------------------------
	MF field may be coded as ListMFDescription or VectorMFDescription
	-----------------------------------------------------------------*/
	if (pBifs -> ParseInt (1))	// isListDescription
		do
		{
			if (!AppendElement (pBifs, pNode, pQuant))
				return FALSE;
			TRACENEWLINE;
		} while (pBifs -> ParseInt (1));
	else
		for (int i = pBifs -> ParseInt (pBifs -> ParseInt (5)); --i >= 0;)
		{
			if (!AppendElement (pBifs, pNode, pQuant))
				return FALSE;
			TRACENEWLINE;
		}

	TRACEINDENT (-1);
	TRACEF ("]", 0);
	TRACENEWLINE;
		
	return TRUE;
}

/*-----------------------------------------------------------------
MFGenericNode destructor deletes all nodes
-----------------------------------------------------------------*/
MFGenericNode :: ~MFGenericNode ()
{
	for (int i = GetLength (); --i >= 0;)
		if (m_array [i])
			m_array [i] -> RemoveContainer (this, m_pParent);
}

/*-----------------------------------------------------------------
Manipulation of MFGenericNode arrays
-----------------------------------------------------------------*/
void MFGenericNode :: InsertAt (int index, MediaObject *value, BOOL bRoute)
{
	m_array.InsertAt (index, value);
	value -> SetContainer (this, TRUE, m_pParent);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: Append (MediaObject *value, BOOL bRoute)
{
	m_array += value;
	value -> SetContainer (this, TRUE, m_pParent);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: SetAt (int index, MediaObject *value, BOOL bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.SetAt (index, value);
	value -> SetContainer (this, TRUE, m_pParent);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: DeleteFrom (int index, BOOL bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.DeleteFrom (index);
	if (bRoute)
		OnFieldUpdate ();
}

/*-----------------------------------------------------------------
Assigment operator activated in case of Routes. We assume that
the field argument is always of the same type as this, therefor
allow for deliberate downcasting. This assumption is dangerous.
Commercial application should implement safe casting or proper
conversion mechanism.
-----------------------------------------------------------------*/
void MFGenericNode :: operator = (NodeField &field)
{
	MFGenericNode &fld = (MFGenericNode &) field;
	for (int i = GetLength (); --i >= 0;)
		DeleteFrom (i, FALSE);		
	for (i = 0; i < fld.GetLength (); i++)
		Append (fld [i], FALSE);
	OnFieldUpdate ();
}

void SFGenericNode::ReplaceNode (MediaObject * /*pNew*/)
{

}

void MFGenericNode::ReplaceNode (MediaObject * /*pOld*/, MediaObject * /*pNew*/)
{

}
