// -*- c++ -*-

#if !defined(_NodeFld_h)	// Sentry, use file only if it's not already included.
#define _NodeFld_h

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       NodeFld.h
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Declarations of NodeField and derivations.
-------------------------------------------------------------------------*/
#include "ztl.h"
#include "ztypes.h"
#include "MediaObj.h"

class BIFSDecoder;
class MediaObject;
class Quantizer;
class Route;

#ifdef BIFSENC
class BIFSEncoder;
#endif

/*-------------------------------------------------------------------------
- NodeField is the base class all kind of node fields
- NodeField objects can parse themselves from BIFS.
- Each object must override the Parse virtual function.
-------------------------------------------------------------------------*/
class NodeField
{
	Route *m_pFirstRoute;

public:

#ifndef BIFSENC
	NodeField () : m_pFirstRoute (NULL) {}
	virtual ~NodeField ();
	virtual BOOL Parse (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	) = 0;
	virtual void Dequantize (Quantizer * /*pQuant*/) {}
	virtual void Clip (Quantizer * /*pQuant*/) {}
	virtual void operator = (NodeField &field) = 0;
	void LinkRoute (Route *pRoute);
	void UnlinkRoute (Route *pRoute);
	void OnFieldUpdate ();

#else
public:
	virtual BOOL Encode (
		BIFSEncoder *pBifs, 
		Quantizer *pQuant = NULL
	) = 0;
	virtual BOOL EncodeElement (BIFSEncoder *pBifs, Quantizer *pQuant = NULL);
#endif // BIFSENC
};

/*-----------------------------------------------------------------
A class template for all SF fields, except of SFNode (or specific type
SFNode fields).  T can be the classes Int, Float, Bool, Color, Time,
Rotation, ZString, Vec2f and Vec3f which are defined in ZTL. 
E.g. SFInt  would be declared as SF<int>, and SFRotation as
SF<Rotation>.
-----------------------------------------------------------------*/
template <class T> class SF : public NodeField
{
	/*-----------------------------------------------------------------
	Assigment operator activated in case of Routes. We assume that
	the field argument is always of the same type as this, therefor
	allow for deliberate downcasting. This assumption is dangerous.
	Commercial application should implement safe casting or proper
	conversion mechanism.
	-----------------------------------------------------------------*/
	virtual void operator = (NodeField &field)
	{
		*this = (T) (SF<T>&) field;		
	}

protected:
	T m_value;

public:

	SF<T> () : m_value (0) {}
	SF<T> (T value) : m_value (value) {}
	operator T () const {return m_value;}

#ifndef BIFSENC
	virtual BOOL Parse (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	);
	T operator = (const T value)
	{
		m_value = value;
		OnFieldUpdate ();
		return value;
	}

#else
	virtual BOOL Encode (
		BIFSEncoder *pBifs, 
		Quantizer *pQuant = NULL
	);
#endif		// BIFSENC
};

/*-----------------------------------------------------------------
SFString is a synonym to SF<ZString>.
-----------------------------------------------------------------*/
class SFString : public SF<ZString>
{
public:
	SFString () {}
	SFString (LPCSTR szString) : SF<ZString> (szString) {}
	operator const ZString & () const {return m_value;}
	const ZString &operator = (const ZString &value)
	{
		SF<ZString> :: operator = (value);
		return value;
	}
	int GetLength () const {return m_value.GetLength ();}
	operator LPCSTR () const {return m_value;}
};

/*-----------------------------------------------------------------
A special case of  SFInt which us used to store objectIDs.  It
inherits all the public methods of SF<int>.
-----------------------------------------------------------------*/
class SFObjectID : public SF<int>
{
#ifndef BIFSENC
	virtual BOOL Parse (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	);
#endif

public:
	SFObjectID (int nID = 1023) : SF<int> (nID) {}
	int operator = (const int value)
	{
		return SF<int> :: operator = (value);
	}


#ifdef BIFSENC
	virtual BOOL Encode (
		BIFSEncoder *pBifs, 
		Quantizer *pQuant = NULL
	);
#endif
};

class BifsFactory
{
public:
	virtual MediaObject *CreateNode () = 0;
};

template <class T> class BifsNodeFactory : public BifsFactory
{
	virtual MediaObject *CreateNode () {return new T;}

public:
	BifsNodeFactory ()
	{
		SFGenericNode :: StoreFactory (T :: nodeCode, this);
	}
};

/*-----------------------------------------------------------------
This is a base class for all fields that are themselves nodes.  Some
fields may contain any node type, so they are defined as SFGenericNode,
but usually only a small amount of node types are allowed for a given
field.  For this purpose, the nodes are classified according to Node
Data Types.  A Node Data Type is implemented by an object derived from
SFGenericNode.
-----------------------------------------------------------------*/
class SFGenericNode : public NodeField
{
protected:
	MediaObject *m_pNode;

public:
	virtual MediaObject *CreateNode (int nodeType);
	static void StoreFactory (int nCode, BifsFactory *pFactory);
	operator MediaObject * () const {return m_pNode;}
	MediaObject *operator -> () const {return m_pNode;}
	MediaObject *operator = (MediaObject *node);

private:
	static BifsFactory *m_bifsFactory [];
	virtual int GetNodeTableSize ();

#ifndef BIFSENC
	/*-----------------------------------------------------------------
	Assigment operator activated in case of Routes. We assume that
	the field argument is always of the same type as this, therefor
	allow for deliberate downcasting. This assumption is dangerous.
	Commercial application should implement safe casting or proper
	conversion mechanism.
	-----------------------------------------------------------------*/
public:
	SFGenericNode () : m_pNode (NULL) {}
	virtual ~SFGenericNode ();
	virtual void operator = (NodeField &field)
	{
		*this = (MediaObject *) (SFGenericNode&) field;
	}

public:
	void ReplaceNode (MediaObject *pNew);
	virtual BOOL Parse (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	);
	void RemoveNode () {m_pNode = 0;}

#else
public:
	virtual BOOL Encode (
		BIFSEncoder *pBifs, 
		Quantizer *pQuant = NULL
	);

protected:
	virtual MediaObject *CreateNode (LPCSTR szNodeName, int &rNodeType);

#endif	// BIFSENC
};

/*-----------------------------------------------------------------
A class template for all fields derived from SFGenericNode, when the
node is of a single known type.  E.g. SFAppearance2Dnode would be
declared as SFNode<Appearance2D>.
-----------------------------------------------------------------*/
template <class T> class SFNode : public SFGenericNode
{
private:
	virtual MediaObject *CreateNode (int);
	virtual int GetNodeTableSize () {return 1;}

public:
	T* operator -> () const {return (T*) m_pNode;}
	operator T* () const {return (T*) m_pNode;}
	MediaObject *operator = (T *node)
	{
		return SFGenericNode :: operator = ((MediaObject *) node);
	}

#ifdef BIFSENC
private:
	virtual MediaObject *CreateNode (LPCSTR szNodeName, int &rNodeType);
#endif
};

/*-----------------------------------------------------------------
For every Node Data Type which includes more than one nodeType,
an SF...Node class is defined.
-----------------------------------------------------------------*/
typedef enum
{
	NT_2D, NT_3D, NT_GEOMETRY2D, NT_GEOMETRY, NT_LAYER, NT_ROOT,
	NT_AUDIO, NT_BEHAVIOR, NT_BEHAVIOR2D, NT_TEXTURE
} NodeType;

template <NodeType type> class SFNodeType : public SFGenericNode
{
private:
	static const int m_nodeTable [];
	virtual MediaObject *CreateNode (int nodeType)
	{
		if (nodeType < 0)
			return NULL;
		assert (nodeType < GetNodeTableSize ());
		return SFGenericNode :: CreateNode (m_nodeTable [nodeType]);
	}
	virtual int GetNodeTableSize ();

public:
	MediaObject *operator = (MediaObject *node)
	{
		return SFGenericNode :: operator = (node);
	}

#ifdef BIFSENC
private:
	virtual MediaObject *CreateNode (LPCSTR szNodeName, int &rNodeType)
	{
		MediaObject *pNode;

		for (int nCode = 0; nCode < GetNodeTableSize (); nCode ++)
		{	pNode = CreateNode (nCode);
			if (pNode && !stricmp (pNode -> GetName (), szNodeName))
			{	rNodeType = nCode;
				return pNode;
			}
		}
		pNode = SFGenericNode :: CreateNode (szNodeName, rNodeType);
		rNodeType = -1;
		return pNode;
	}
#endif	// BIFSENC
};

typedef SFNodeType<NT_2D> SF2DNode;
typedef SFNodeType<NT_3D> SF3DNode;
typedef SFNodeType<NT_GEOMETRY2D> SFGeometry2DNode;
typedef SFNodeType<NT_GEOMETRY> SFGeometryNode;
typedef SFNodeType<NT_ROOT> SFRootNode;
typedef SFNodeType<NT_AUDIO> SFAudioNode;
typedef SFNodeType<NT_BEHAVIOR> SFBehaviorNode;
typedef SFNodeType<NT_BEHAVIOR2D> SFBehavior2DNode;
typedef SFNodeType<NT_TEXTURE> SFTextureNode;

/*-----------------------------------------------------------------
A base class for all MF... classes
-----------------------------------------------------------------*/
class MField : public NodeField
{
public:
	virtual int GetLength () const = 0;

#ifndef BIFSENC
	virtual BOOL Parse (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	);
	virtual void DeleteFrom (int index, BOOL bRoute = TRUE) = 0;
	virtual BOOL AppendElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	) = 0;
	virtual BOOL InsertElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	) = 0;
	virtual BOOL ReplaceElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	) = 0;

#else
private:
	virtual BOOL Encode (
		BIFSEncoder *pBifs, 
		Quantizer *pQuant = NULL
	);
#endif		// BIFSINC
};

/*-----------------------------------------------------------------
A class template for all MF fields, except of MFNode (or specific type
MFNode fields).  T can be int, float, BOOL, or the classes as Color,
Time, Rotation, ZString, Vec2f and Vec3f which are defined in ZTL.
-----------------------------------------------------------------*/
template <class T> class MF : public MField
{
protected:
	ZArray<T> m_array;

public:
	const T &operator [] (int index) {return m_array [index];}
	int GetLength () const {return m_array.GetLength ();}
	T operator += (T t) {return m_array += t;}

#ifndef BIFSENC
private:
	virtual BOOL AppendElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	)
	{
		SF<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		m_array += m_element;
		return TRUE;
	}

	virtual BOOL InsertElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		SF<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		m_array.InsertAt (nIndex, m_element);
		return TRUE;
	}

	virtual BOOL ReplaceElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		SF<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		m_array.SetAt (nIndex, m_element);
		return TRUE;
	}

	/*-----------------------------------------------------------------
	Assigment operator activated in case of Routes. We assume that
	the field argument is always of the same type as this, therefor
	allow for deliberate downcasting. This assumption is dangerous.
	Commercial application should implement safe casting or proper
	conversion mechanism.
	-----------------------------------------------------------------*/
	virtual void operator = (NodeField &field)
	{
		MF<T> &fld = (MF<T> &) field;
		m_array.Clear ();
		for (int i = 0; i < fld.GetLength (); i++)
			m_array += fld [i];
		OnFieldUpdate ();
	}

public:
	void InsertAt (int index, T value, BOOL bRoute = TRUE)
	{
		m_array.InsertAt (index, value);
		if (bRoute)
			OnFieldUpdate ();
	}
	void Append (T value, BOOL bRoute = TRUE)
	{
		m_array += value;
		if (bRoute)
			OnFieldUpdate ();
	}
	void SetAt (int index, T value, BOOL bRoute = TRUE)
	{
		m_array.SetAt (index, value);
		if (bRoute)
			OnFieldUpdate ();
	}
	void DeleteFrom (int index, BOOL bRoute = TRUE)
	{
		m_array.DeleteFrom (index);
		if (bRoute)
			OnFieldUpdate ();
	}

#else
private:
	virtual BOOL EncodeElement (BIFSEncoder *pBifs, Quantizer *pQuant = NULL)
	{
		SF<T> m_element;
		return m_element.Encode (pBifs, pQuant);
	}
#endif		// BIFSENC
};

/*-----------------------------------------------------------------
A base class for classes implementing MFNode (or specific type
MFNode fields).
-----------------------------------------------------------------*/
class MFGenericNode : public MF<MediaObject*>
{
	GroupNode *m_pParent;

public:
	MediaObject *operator [] (int index) {return m_array [index];}

#ifndef BIFSENC
private:
	virtual void operator = (NodeField &field);

public:
	virtual ~MFGenericNode ();
	void ReplaceNode (MediaObject *pOld, MediaObject *pNew);
	MFGenericNode () {m_pParent = NULL;}
	void RemoveNode (MediaObject *pNode) {m_array -= pNode;}
	void SetParent (GroupNode *pNode) {m_pParent = pNode;}
	void InsertAt (int index, MediaObject * value, BOOL bRoute = TRUE);
	void Append (MediaObject * value, BOOL bRoute = TRUE);
	void SetAt (int index, MediaObject * value, BOOL bRoute = TRUE);
	void DeleteFrom (int index, BOOL bRoute = TRUE);
	virtual BOOL AppendElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	)
	{
		SFGenericNode m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		Append (m_element, FALSE);
		return TRUE;
	}

	virtual BOOL InsertElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode,
		int nIndex,	 
		Quantizer *pQuant = NULL
	)
	{
		SFGenericNode m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		InsertAt (nIndex, m_element, FALSE);
		return TRUE;
	}

	virtual BOOL ReplaceElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode,
		int nIndex,	 
		Quantizer *pQuant = NULL
	)
	{
		SFGenericNode m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		SetAt (nIndex, m_element, FALSE);
		return TRUE;
	}

#else
private:
	virtual BOOL EncodeElement (BIFSEncoder *pBifs, Quantizer *pQuant = NULL)
	{
		SFGenericNode m_element;
		return m_element.Encode (pBifs, pQuant);
	}
#endif		// BIFSENC
};

/*-----------------------------------------------------------------
A class template for all MFNode fields which allow nodes of a single
known type.  E.g. MFAppearance2Dnode would be declared as
MFNode<Appearance2D>.
-----------------------------------------------------------------*/
template <class T> class MFNode : public MFGenericNode
{
public:
	T *operator [] (int index) {return (T*) m_array [index];}

#ifndef BIFSENC
private:
	virtual BOOL AppendElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	)
	{
		SFNode<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		Append (m_element, FALSE);
		return TRUE;
	}

	virtual BOOL InsertElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		SFNode<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		InsertAt (nIndex, m_element, FALSE);
		return TRUE;
	}

	virtual BOOL ReplaceElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		SFNode<T> m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		SetAt (nIndex, m_element, FALSE);
		return TRUE;
	}

#else
private:
	virtual BOOL EncodeElement (BIFSEncoder *pBifs, Quantizer *pQuant = NULL)
	{
		SFNode<T> m_element;
		return m_element.Encode (pBifs, pQuant);
	}
#endif		// BIFSENC
};

/*-----------------------------------------------------------------
A class template for all MFNode fields which allow more than one node
type.  T can be any of the SF...Node types. E.g. MF2DNode would be
declared as MFNodeType<SF2DNode>.
-----------------------------------------------------------------*/
template <class T> class MFNodeType : public MFGenericNode
{
#ifndef BIFSENC
private:
	virtual BOOL AppendElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		Quantizer *pQuant = NULL
	)
	{
		T m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		Append (m_element, FALSE);
		return TRUE;
	}

	virtual BOOL InsertElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		T m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		InsertAt (nIndex, m_element, FALSE);
		return TRUE;
	}

	virtual BOOL ReplaceElement (
		BIFSDecoder *pBifs, 
		MediaObject *pNode, 
		int nIndex,
		Quantizer *pQuant = NULL
	)
	{
		T m_element;
		if (!m_element.Parse (pBifs, pNode, pQuant))
			return FALSE;
		SetAt (nIndex, m_element, FALSE);
		return TRUE;
	}

#else
private:
	virtual BOOL EncodeElement (BIFSEncoder *pBifs, Quantizer *pQuant = NULL)
	{
		T m_element;
		return m_element.Encode (pBifs, pQuant);
	}
#endif		// BIFSENC
};

#endif                                      // _NodeFld_h sentry.
