/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Buffer management
FILE:       Stream.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of MediaStream.
			(This version simulation only. Just makes sure Allocate,
			Reallocate, and Dispatch are called properly.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "Im1defs.h"
#include "Clock.h"
#include "trace.h"

int MediaStream :: m_nStreamCount;
FILE * MediaStream :: m_fTraceFile;
ZMutex MediaStream :: m_mtxLockTraceFile;

/*-----------------------------------------------------------------
MediaStream constructor. The simulation allocate a buffer large enough 
for all possible AUs
-----------------------------------------------------------------*/
MediaStream :: MediaStream (ClockReference *pClock, int nBufferSize) :
	m_smpWaitForMemory (1, 0),
	m_smpWaitForData (1, 0)
{
	/*-----------------------------------------------------------------
	Allocate the stream buffer.
	-----------------------------------------------------------------*/
	m_nTakenStart = m_nBufferSize = nBufferSize;
	m_pBuffer = (LPBYTE) malloc (nBufferSize);
	m_nSN = m_nStreamCount++;		// For tracing
	m_pClock = pClock;
	m_bEndOfStream = TRUE;
	Reset ();
}

/*-----------------------------------------------------------------
Reset the stream
-----------------------------------------------------------------*/
void MediaStream :: Reset ()
{
	/*-----------------------------------------------------------------
	two semaphores used to make sure a stream in use is not deleted
	-----------------------------------------------------------------*/
	if (m_bEndOfStream)
	{	m_smpReceiveClosed.P ();
		m_smpSendClosed.P ();
	}

	m_nFreeStart = 0;
	m_nAllocatedLength = m_nLockedLength = -1;
	m_bWaitingForMemory = m_bReallocInProgress = m_bEndOfStream = FALSE;
	m_bFirstRetrieved = FALSE;
	m_bTrace = FALSE;
}

/*-----------------------------------------------------------------
MediaStream destructor
-----------------------------------------------------------------*/
MediaStream :: ~MediaStream ()
{
	/*-----------------------------------------------------------------
	Wait till both sides close the stream
	-----------------------------------------------------------------*/
	m_smpReceiveClosed.P ();
	m_smpSendClosed.P ();
	free (m_pBuffer);

	/*-----------------------------------------------------------------
	Following lines for tracing
	-----------------------------------------------------------------*/
	if (--m_nStreamCount == 0 && m_fTraceFile != NULL)
	{	fclose (m_fTraceFile);
		m_fTraceFile = NULL;
	}
}

/*-----------------------------------------------------------------
Allocate() allocates a block of nSize byte and returns its address
-----------------------------------------------------------------*/
LPVOID MediaStream :: Allocate (int nSize, BOOL bWait, BOOL bShove)
{
	assert (nSize > 0);

	/*-----------------------------------------------------------------
	Return NULL if last allocated block has not been dispatched
	or if required size exceeds buffer size
	-----------------------------------------------------------------*/
	nSize += sizeof (int) + sizeof (DWORD);	// for block length and time stamp
	if (m_nAllocatedLength >= 0 && !m_bReallocInProgress ||
		nSize > m_nBufferSize
	)
		return NULL;

	/*-----------------------------------------------------------------
	Check available (contiguous) space.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	if (m_bEndOfStream)
	{	m_mtxLockStream.V ();
		if (!m_bReallocInProgress)
			m_smpSendClosed.V ();
		return END_OF_STREAM;
	}
	while (
		nSize > m_nTakenStart - m_nFreeStart &&
		(	m_nTakenStart > m_nFreeStart ||
			m_nTakenStart == m_nFreeStart && !m_bReallocInProgress ||
			nSize > m_nBufferSize - m_nFreeStart &&
			nSize > m_nTakenStart
		)
	)
	{
		/*-----------------------------------------------------------------
		Not enough space available. One reason might be that the buffer
		is empty, and this is a reallocation of a block which is
		currently at the middle of the buffer
		-----------------------------------------------------------------*/
		if (m_nTakenStart == m_nFreeStart && m_bReallocInProgress)
		{	m_nFreeStart = 0;
			m_nTakenStart = m_nBufferSize;
		}
		/*-----------------------------------------------------------------
		If bShove it TRUE, shove last block
		(or wait for the currently locked block to be discarded).
		-----------------------------------------------------------------*/
		else if (bShove)
		{	if (m_nLockedLength >= 0)
			{	m_mtxLockStream.V ();
				m_smpLockOutBlock.P ();
				m_smpLockOutBlock.V ();
				m_mtxLockStream.P ();
			}
			else
				Discard ();
		}
		/*-----------------------------------------------------------------
		If the wait flag is true, wait till a space is available
		-----------------------------------------------------------------*/
		else if (bWait)
		{	m_bWaitingForMemory = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForMemory.P ();
			m_mtxLockStream.P ();
			if (m_bEndOfStream)
			{	m_mtxLockStream.V ();
				if (!m_bReallocInProgress)
					m_smpSendClosed.V ();
				return END_OF_STREAM;
			}
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Allocate a block of the desired length. Starting either at the next
	available byte, or cycle around the buffer
	-----------------------------------------------------------------*/
	if (m_nBufferSize - m_nFreeStart < nSize)
	{	STOREINT (-1, m_pBuffer + m_nFreeStart);	// Mark as not used
		if (m_nTakenStart == m_nFreeStart)
			m_nTakenStart = m_nBufferSize;
		m_nFreeStart = 0;
	}
	m_nAllocatedLength = nSize - (sizeof (int) + sizeof (DWORD));
	m_bWait = bWait;			// save for realloc
	m_bShove = bShove;
	m_mtxLockStream.V ();
	return m_pBuffer + m_nFreeStart + sizeof (int) + sizeof (DWORD);
}

/*-----------------------------------------------------------------
Reallocate() expands the reserved block, if necessary
-----------------------------------------------------------------*/
LPVOID MediaStream :: Reallocate (int nSize)
{
	assert (nSize >= 0);

	/*-----------------------------------------------------------------
	Reallocation is done by zeroing m_nAllocatedLength and calling
	Allocate with the new size.
	-----------------------------------------------------------------*/

	m_mtxLockStream.P ();

	int nPreviousStart = m_nFreeStart;
	int nPreviousLength = m_nAllocatedLength;
	LPVOID pResult;
	
	m_bReallocInProgress = TRUE;
	m_mtxLockStream.V ();
	pResult = Allocate (nSize, m_bWait, m_bShove);
	m_bReallocInProgress = FALSE;

	/*-----------------------------------------------------------------
	If the stream was closed by the receiving side, tell the receiver
	not to wait till block is discarded, and return END_OF_STREAM
	-----------------------------------------------------------------*/
	if (pResult == END_OF_STREAM)
	{	m_nAllocatedLength = -1;
		if (m_bWaitingForData)
		{	m_bWaitingForData = FALSE;
			m_smpWaitForData.V ();
			m_smpSendClosed.V ();
		}
		return END_OF_STREAM;
	}

	/*-----------------------------------------------------------------
	If realloc successful, but block was shifted in memory, copy the
	contents of the old block.
	-----------------------------------------------------------------*/
	if (pResult != NULL && m_nFreeStart != nPreviousStart)
		memcpy (
			pResult,
			m_pBuffer + nPreviousStart + sizeof (int) + sizeof (DWORD),
			nPreviousLength
		);
	return pResult;
}

/*-----------------------------------------------------------------
Dispatch() dispatches the allocated block into the stream
-----------------------------------------------------------------*/
BOOL MediaStream :: Dispatch (int nSize, DWORD dwTime)
{
	assert (nSize >= 0);

	if (m_nAllocatedLength < 0)
		return FALSE;

	/*-----------------------------------------------------------------
	store the size and time information. Signal to waiting threads.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	nSize += sizeof (int) + sizeof (DWORD);
	STOREINT (nSize, m_pBuffer + m_nFreeStart);
	STOREDWORD (dwTime, m_pBuffer + m_nFreeStart + sizeof (int));

	Trace (m_nFreeStart, "Dispatch");

	m_nFreeStart += nSize;
	if (m_nFreeStart == m_nBufferSize)
		m_nFreeStart = 0;
	if (m_nTakenStart == m_nBufferSize)
		m_nTakenStart = 0;
	m_nAllocatedLength = -1;
	if (m_bWaitingForData)
	{	m_bWaitingForData = FALSE;
		m_smpWaitForData.V ();
	}
	m_mtxLockStream.V ();
	return TRUE;
}

/*-----------------------------------------------------------------
FetchNext fetches the next available block, ignoring its time stamp (which
is only returned in rdwTime.  If bWait is true, the function waits till
a block becomes available.
If successful, the function locks the fetched block.
If end of stream reached, the function returns END_OF_STREAM.
-----------------------------------------------------------------*/
LPVOID MediaStream :: FetchNext (
	int &rnSize,
	DWORD &rdwTime,
	BOOL bWait
)
{
	/*-----------------------------------------------------------------
	Returns NULL if a block is already locked
	-----------------------------------------------------------------*/
	if (m_nLockedLength >= 0)
		return NULL;

	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	while (
		m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			m_smpReceiveClosed.V ();
			return END_OF_STREAM;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}
	
	/*-----------------------------------------------------------------
	Lock a block and return its address, size and time stamp
	-----------------------------------------------------------------*/
	m_smpLockOutBlock.P ();
	m_nLockedLength = GetBlockLength (m_nTakenStart);
	rnSize = m_nLockedLength - sizeof (int) - sizeof (DWORD);
	rdwTime = GetTimeStamp (m_nTakenStart);
	Trace (m_nTakenStart, "Fetch next");
	m_mtxLockStream.V ();
	return m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
}

/*-----------------------------------------------------------------
Fetch() fetches a block whose time stamp matured.
It will return TRUE rbRecurrent if it's the same block returned on
previous call.
If bSkip is TRUE it will skip all blocks till the newest one whose time
matches.
If bWait is TRUE it will wait till one block matures.
-----------------------------------------------------------------*/
LPVOID MediaStream :: Fetch (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &rbRecurrent,
	BOOL &rbEndOfStream,
	BOOL bSkip,
	BOOL bWait,
	long lDelta		//ML changed from DWORD to long
)
{
	rbEndOfStream = FALSE;

	/*-----------------------------------------------------------------
	Returns NULL if a block is already locked
	-----------------------------------------------------------------*/
	if (m_nLockedLength >= 0)
		return NULL;

	/*-----------------------------------------------------------------
	If first block has been already retrieved and bWait is true, discard
	that block.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	if (m_bFirstRetrieved && bWait)
		Discard ();

	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	while (
		m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			//m_smpReceiveClosed.V ();
			rbEndOfStream = TRUE;
			return NULL;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Find the next block which matured. If bSkip is TRUE, find the newest
	matured block.
	-----------------------------------------------------------------*/
	int nNext = m_nTakenStart;
	int i = nNext;
	DWORD dwTime;
	DWORD dwCurrentTime;
	if ((long)m_pClock -> GetTime () + lDelta < 0)
		dwCurrentTime = 0;
	else
		dwCurrentTime = m_pClock -> GetTime () + lDelta;

	TRACE_PRINT3("MediaStream: Current time = %d (%d + %d)\n",
				 dwCurrentTime, m_pClock->GetTime(), lDelta);

	if (m_bFirstRetrieved)
		i = GetNext (nNext);
	while (
		i >= 0 &&
		(dwTime = GetTimeStamp (i)) != 0 &&
		m_pClock -> CompareTimes (dwCurrentTime, dwTime) >= 0
	)
	{	nNext = i;
		if (!bSkip)
			break;
		i = GetNext (nNext);
	}

	/*-----------------------------------------------------------------
	Now nNext points to the most suitable block.
	If it is a matured block, lock it and return its address.
	Otherwise wait or return NULL
	-----------------------------------------------------------------*/
	dwTime = GetTimeStamp (nNext);
	if (m_pClock -> CompareTimes (dwTime, dwCurrentTime) > 0)
	{	m_mtxLockStream.V ();
		if (bWait)
		{	m_pClock -> WaitOn (dwTime);
			m_mtxLockStream.P ();
		}
		else
			return NULL;
	}

	/*-----------------------------------------------------------------
	Now we can lock the block and return all the required information
	-----------------------------------------------------------------*/
	m_smpLockOutBlock.P ();
	m_nLockedLength = GetBlockLength (nNext);
	rnSize = m_nLockedLength - sizeof (int) - sizeof (DWORD);
	rdwTime = GetTimeStamp (m_nTakenStart);
	rbRecurrent = nNext == m_nTakenStart && m_bFirstRetrieved;
	m_bFirstRetrieved = TRUE;
	Trace (nNext, "Fetch");

	/*-----------------------------------------------------------------
	Discard old blocks
	-----------------------------------------------------------------*/
	if (nNext != m_nTakenStart)
	{	m_nTakenStart = nNext;
		if (m_bWaitingForMemory)
		{	m_bWaitingForMemory = FALSE;
			m_smpWaitForMemory.V ();
		}
	}			
	m_mtxLockStream.V ();
	if (rbRecurrent && m_bEndOfStream)
	{
		//ML added this next test
		// Only indicate that the end of the stream is reached when
		// there are no un-read units in the buffer
		if (i < 0)
		{
			//m_smpReceiveClosed.V ();
			rbEndOfStream = TRUE;
		}
	}
	return m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
}

/*-----------------------------------------------------------------
Releases a previously locked output block. If bDiscard is TRUE, or its
time stamp is zero, discard the block from the stream
-----------------------------------------------------------------*/
void MediaStream :: Release (BOOL bDiscard)
{
	/*-----------------------------------------------------------------
	Do nothing if no block is locked
	-----------------------------------------------------------------*/
	if (m_nLockedLength < 0)
		return;

	m_mtxLockStream.P ();
	m_nLockedLength = -1;
	m_smpLockOutBlock.V ();
	if (bDiscard)
		Discard ();
	m_mtxLockStream.V ();
}

/*-----------------------------------------------------------------
Mark that input side of the stream is closed.
-----------------------------------------------------------------*/
void MediaStream :: Close ()
{
	m_mtxLockStream.P ();
	m_bEndOfStream = TRUE;
	if (m_bWaitingForData)
	{	m_bWaitingForData = FALSE;
		m_smpWaitForData.V ();
	}
	m_mtxLockStream.V ();
	m_smpSendClosed.V ();
}
	
/*-----------------------------------------------------------------
Discard next block
-----------------------------------------------------------------*/
void MediaStream :: Discard ()
{
	m_nTakenStart += GetBlockLength (m_nTakenStart);
	if (m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart != m_nFreeStart &&
		GetBlockLength (m_nTakenStart) == -1
	)
		m_nTakenStart = 0;

	/*-----------------------------------------------------------------
	If the buffer is now empty and no block is currently
	allocated, we can reset the TakenStart and FreeStart values
	-----------------------------------------------------------------*/
	if (m_nTakenStart == m_nFreeStart && m_nAllocatedLength < 0)
	{	m_nFreeStart = 0;
		m_nTakenStart = m_nBufferSize;
	}
	m_bFirstRetrieved = FALSE;
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = FALSE;
		m_smpWaitForMemory.V ();
	}
}

/*-----------------------------------------------------------------
Find the offset of the block following the one whose offset is given
Returns -1 if there are no more available blocks.
-----------------------------------------------------------------*/
int MediaStream :: GetNext (int nOffset)
{
	int i = GetBlockLength (nOffset);

	if (i == -1)
		nOffset = 0;
	else
		nOffset += i;
	if (nOffset >= m_nBufferSize)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	if (GetBlockLength (nOffset) == -1)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	return nOffset;
}

/*-----------------------------------------------------------------
Trace a stream action into a file
-----------------------------------------------------------------*/
void MediaStream :: Trace (int nOffset, LPCSTR szAction)
{
	if (!m_bTrace)
		return;

	m_mtxLockTraceFile.P ();
	if (m_fTraceFile == NULL)
		m_fTraceFile = fopen ("Trace.txt", "w");
	fprintf (
		m_fTraceFile,
		"Time: %d, stream #: %d, action: %s, size: %d, time-stamp: %d\n",
		m_pClock -> GetTime (),
		m_nSN,
		szAction,
		GetBlockLength (nOffset),
		GetTimeStamp (nOffset)
	);
	m_mtxLockTraceFile.V ();
}


/*-----------------------------------------------------------------
Close the stream by the receiving peer
-----------------------------------------------------------------*/
void MediaStream :: Terminate ()
{
	Release ();
	m_mtxLockStream.P ();
	m_bEndOfStream = TRUE;

	/*-----------------------------------------------------------------
	If the sender waits for memory tell it to go
	-----------------------------------------------------------------*/
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = FALSE;
		m_smpWaitForMemory.V ();
	}

	m_nAllocatedLength = -1;
	m_mtxLockStream.V ();
	m_smpReceiveClosed.V ();
}
