/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Multiplexer
FILE:       MuxOD.cpp
AUTHOR:     Zvi Lifshitz (VDOnet Corp.)
OVERVIEW:   Implementation of MuxObjectDescriptor
-------------------------------------------------------------------------*/
#include "defs.h"
#include "MuxOD.h"
#include "Token.h"
#include "Stream.h"
#include "ESConfig.h"

enum {CODE_MuxObjectDescriptor};

BIFS_IMPLEMENT_NODE_START(MuxObjectDescriptor)
	 BIFS_FIELD(MuxObjectDescriptor, objectDescriptorID, 0, 0, -1, -1, 0, 0)
	 BIFS_FIELD(MuxObjectDescriptor, decTypeString, 1, 1, -1, -1, 0, 0)
	 BIFS_FIELD1(MuxObjectDescriptor, configParam, 2, 2, -1, -1, 0, 0, 0, 127)
	 BIFS_FIELD(MuxObjectDescriptor, startTime, 3, 3, -1, -1, 0, 0)
	 BIFS_FIELD(MuxObjectDescriptor, fileName, 4, 4, -1, -1, 0, 0)
	 BIFS_FIELD(MuxObjectDescriptor, streamData, 4, 4, -1, -1, 0, 0)
BIFS_IMPLEMENT_NODE_END(MuxObjectDescriptor)

/*-----------------------------------------------------------------
MuxObjectDescriptor destructor. Close the stream file
-----------------------------------------------------------------*/
MuxObjectDescriptor :: ~MuxObjectDescriptor ()
{
	delete m_pStream;
	if (m_fStream)
		fclose (m_fStream);
}

/*-----------------------------------------------------------------
Array of factory objects to create proper Stream object
-----------------------------------------------------------------*/
StreamFactoryBase *MuxObjectDescriptor :: m_pFactories [40];
int MuxObjectDescriptor :: m_nFactoriesAvailable;

/*-----------------------------------------------------------------
Register a StreamFactory object
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: RegisterFactory (StreamFactoryBase *pFactory)
{
	assert (m_nFactoriesAvailable < 40);
	m_pFactories [m_nFactoriesAvailable++] = pFactory;
}

/*-----------------------------------------------------------------
Parse an ObjectDescriptor, open the stream file, and create a proper
Stream object
-----------------------------------------------------------------*/
BOOL MuxObjectDescriptor :: Parse (FILE *fMP4, int nFMC)
{
	m_bLastByte = 0;
	m_nBitCount = 0;
	m_fMP4 = fMP4;
	m_nFMC = nFMC;
	m_lTS = -1;

	/*-----------------------------------------------------------------
	Make sure the script file contains an object descriptor
	-----------------------------------------------------------------*/
	if (nextToken != TID_NAME ||
		stricmp (nextToken.GetName (), "ObjectDescriptor")
	)
	{	nextToken.Error ("Unexpected symbol");
		nextToken.Scan ();
		return FALSE;
	}

	nextToken.Scan ();
	if (!Encode (NULL))
	{	printf ("Script file parsing error\n");
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Open the stream file
	-----------------------------------------------------------------*/
	m_fStream = fopen (fileName, "rb");
	if (!m_fStream)
	{	printf ("Cannot open %s\n", (LPCSTR) fileName);
		return FALSE;
	}

	/*-----------------------------------------------------------------
	Assign ConfigParams
	-----------------------------------------------------------------*/
	extern ESCONFIGPARAMS g_arPredefined [];
	extern int g_nPredefinedLength;

	if (configParam < 0 || configParam >= g_nPredefinedLength)
	{	printf ("Invalid ConfigParams: %d\n", (int) configParam);
		return FALSE;
	}
	m_pConfigParams = g_arPredefined + configParam;

	/*-----------------------------------------------------------------
	Instantiate a proper Stream object
	-----------------------------------------------------------------*/
	for (int i = m_nFactoriesAvailable; --i >= 0;)
		if (m_pFactories [i] -> IsTypeOf (decTypeString))
		{	m_pStream = m_pFactories [i] -> Create (m_fStream, streamData);
			break;
		}
	if (!m_pStream)
	{	printf ("Unknown decoder type: %s\n", (LPCSTR) decTypeString);
		return FALSE;
	}

	m_nSequenceNumber = 0;
	return TRUE;
}

/*-----------------------------------------------------------------
Ask the stream for next time stamp
-----------------------------------------------------------------*/
long MuxObjectDescriptor :: GetTimeStamp ()
{
	if (m_lTS < 0)
		m_lTS = m_pStream -> GetTimeStamp ();
	return m_lTS;
}

/*-----------------------------------------------------------------
Store one access unit
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: Store ()
{
	int nPayloadLength = 0;
	BYTE payload [255];
	BOOL bStart = TRUE;

	for (;;)
	{	int c = m_pStream -> GetChar ();
		if (c < 0 || nPayloadLength == 255)
		{
			/*-----------------------------------------------------------------
			Prepare the header. If after that it turns out that the
			header + payload is too long, truncate the payload (and make
			sure the header is not for an "end" payload
			-----------------------------------------------------------------*/
			BOOL bEnd = c < 0;
			PrepareHeader (bStart, bEnd);
			if (c < 0 && m_nHeaderLength + nPayloadLength > 255)
				PrepareHeader (bStart, bEnd = FALSE);

			/*-----------------------------------------------------------------
			Store the packet. If this is end of AU, store the rest of
			the AU in another packet (if necessary), and break the loop.
			-----------------------------------------------------------------*/
			int nPacketLength = 255 - m_nHeaderLength;
			if (nPacketLength > nPayloadLength)
				nPacketLength = nPayloadLength;
			StorePacket (payload, nPacketLength);
			if (c < 0)
			{	if (!bEnd)
				{	PrepareHeader (FALSE, TRUE);
					StorePacket (
						payload + nPacketLength,
						nPayloadLength - nPacketLength
					);
				}
				break;
			}
				
			/*-----------------------------------------------------------------
			The data that was not stored is shifted to the beginning of
			the buffer
			-----------------------------------------------------------------*/
			nPayloadLength -= nPacketLength;
			if (nPayloadLength > 0)
				memcpy (payload, payload + nPacketLength, nPayloadLength);
			bStart = FALSE;
		}
		payload [nPayloadLength++] = c;
	}
	m_lTS = -1;
}

/*-----------------------------------------------------------------
Prepare a packet header
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: PrepareHeader (BOOL bStart, BOOL bEnd)
{
	m_nHeaderLength = 0;
	m_bLastByte = 0;
	m_nBitCount = 0;
	if (m_pConfigParams -> m_bUseAccessUnitStartFlag)
		StoreHeaderBits (bStart, 1);
	if (m_pConfigParams -> m_UseRandomAccessPointFlag)
		StoreHeaderBits (0, 1);
	if (m_pConfigParams -> m_UseAccessUnitEndFlag)
		StoreHeaderBits (bEnd, 1);
	if (m_pConfigParams -> m_nOCRlength > 0)
		StoreHeaderBits (0, 1);
//ML - commented out this line		if (m_pConfigParams -> m_bUsePaddingFlag && bEnd)
	if (m_pConfigParams -> m_bUsePaddingFlag)
	{	StoreHeaderBits (m_pStream -> GetPaddingBits () > 0, 1);
		if (m_pStream -> GetPaddingBits () > 0)
			StoreHeaderBits (m_pStream -> GetPaddingBits (), 3);
	}
	StoreHeaderBits (
		m_nSequenceNumber,
		m_pConfigParams -> m_nSeqNumLength
	);
	if (bStart)
	{
		if (m_pConfigParams -> m_bUseTimeStamps)
		{	StoreHeaderBits (0, 1);
			StoreHeaderBits (1, 1);
			StoreHeaderBits (0, 1);
		}
		if (m_pConfigParams -> m_nAUlength > 0)
			StoreHeaderBits (0, 1);
		if (m_pConfigParams -> m_nInstantBitrateLength > 0)
			StoreHeaderBits (0, 1);
		if (m_pConfigParams -> m_bUseTimeStamps)
			StoreHeaderBits (
				m_lTS *
				(m_pConfigParams -> m_nTimeStampResolution / 1000.),
				m_pConfigParams -> m_nTimeStampLength
			);
		StoreHeaderBits (0, m_pConfigParams -> m_nDegradationPriorityLength);
	}
	FlushHeader ();
}

/*-----------------------------------------------------------------
Store one packet (header + payload) in the muxed file
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: StorePacket (LPBYTE packet, int nPacketLength)
{
	fputc (m_nFMC, m_fMP4);
	fputc (m_nHeaderLength + nPacketLength, m_fMP4);
	for (int i = 0; i < m_nHeaderLength; i++)
		fputc (m_header [i], m_fMP4);
	for (i = 0; i < nPacketLength; i++)
		fputc (packet [i], m_fMP4);
	++m_nSequenceNumber;
}

/*-----------------------------------------------------------------
Store a value as a bit field in the header array
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: StoreHeaderBits (int nValue, int nLength)
{
	nValue <<= sizeof (int) * 8 - nLength;
	while (--nLength >= 0)
	{	m_bLastByte <<= 1;
		m_bLastByte |= (nValue < 0);
		if (++m_nBitCount == 8)
		{	m_header [m_nHeaderLength++] = m_bLastByte;
			m_bLastByte = 0;
			m_nBitCount = 0;
		}
		nValue <<= 1;
	}
}

/*-----------------------------------------------------------------
Add remaining bits to the header
-----------------------------------------------------------------*/
void MuxObjectDescriptor :: FlushHeader ()
{
	if (m_nBitCount > 0)
	{	m_bLastByte <<= 8 - m_nBitCount;
		m_header [m_nHeaderLength++] = m_bLastByte;
	}
}

/*-----------------------------------------------------------------
The StreamFactoryBase constructor registers a factory object
-----------------------------------------------------------------*/
StreamFactoryBase :: StreamFactoryBase (LPCSTR szType)
{
	m_szType = szType;
	MuxObjectDescriptor :: RegisterFactory (this);
}